import argparse
import sys
from pathlib import Path

import yaml

from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.util.PathBuilder import PathBuilder


def build_metadata_column_mapping(columns_str):
    colnames = columns_str.split(",")
    colnames = [label.strip().strip("'\"") for label in colnames]

    return {colname: colname for colname in colnames if colname != ""}

def get_dataset_specs(args):
    dataset_specs = {"format": args.format,
                     "params": {"region_type": RegionType.IMGT_CDR3.name,
                                "result_path": "./",
                                "path": "./"}}

    if args.is_repertoire == "True":
        dataset_specs["params"]["is_repertoire"] = True
        dataset_specs["params"]["metadata_file"] = args.metadata_file
    else:
        dataset_specs["params"]["is_repertoire"] = False

        paired = True if args.paired == "True" else False

        dataset_specs["params"]["paired"] = paired
        if paired:
            dataset_specs["params"]["receptor_chains"] = args.receptor_chains

        if args.metadata_columns != "":
            dataset_specs["params"]["metadata_column_mapping"] = build_metadata_column_mapping(args.metadata_columns)

    return {args.dataset_name: dataset_specs}


def build_specs(args):
    specs = {
        "definitions": {
            "datasets": get_dataset_specs(args),
        },
        "instructions": {
            "my_dataset_generation_instruction": {
                "type": "DatasetExport",
                "datasets": [args.dataset_name],
                "export_formats": ["ImmuneML"]
            }
        }
    }


    return specs


def parse_commandline_arguments(args):
    parser = argparse.ArgumentParser(description="Tool for building immuneML defintion YAML for Galaxy Create Dataset tool")
    parser.add_argument("-r", "--format", required=True, help="The format of the repertoire/receptor dataset")
    parser.add_argument("-m", "--metadata_file", default="",
                        help="The metadata file when using a repertoire dataset. When using a receptor dataset, you may supply an empty string.")
    parser.add_argument("-i", "--is_repertoire", choices=["True", "False"], required=True, help="Whether to import a RepertoireDataset")
    parser.add_argument("-p", "--paired", choices=["True", "False"], default="False",
                        help="When the data is not repertoire data (metadata file = ''), this specifies whether the data is paired (ReceptorDataset) or unpaired (SequenceDataset)")
    parser.add_argument("-c", "--receptor_chains", choices=["TRA_TRB", "TRG_TRD", "IGH_IGL", "IGH_IGK"], default="TRA_TRB",
                        help="When the data is a ReceptorDataset, this specifies the type of receptor chains that are used.")
    parser.add_argument("-a", "--metadata_columns", default="", help="The name of metadata columns of a Sequence- or ReceptorDataset.")
    parser.add_argument("-d", "--dataset_name", default="dataset", help="The name of the created dataset.")

    parser.add_argument("-o", "--output_path", required=True, help="Output location for the generated yaml file (directiory).")
    parser.add_argument("-f", "--file_name", default="specs.yaml",
                        help="Output file name for the yaml file. Default name is 'specs.yaml' if not specified.")

    return parser.parse_args(args)


def main(args):
    parsed_args = parse_commandline_arguments(args)
    specs = build_specs(parsed_args)

    PathBuilder.build(parsed_args.output_path)
    output_location = Path(parsed_args.output_path) / parsed_args.file_name

    with output_location.open("w") as file:
        yaml.dump(specs, file)

    return str(output_location)


if __name__ == "__main__":
    main(sys.argv[1:])



import logging
from pathlib import Path
from typing import List, Tuple

from immuneML.data_model.dataset.ReceptorDataset import ReceptorDataset
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.hyperparameter_optimization.HPSetting import HPSetting
from immuneML.ml_methods.classifiers.MLMethod import MLMethod
from immuneML.ml_methods.classifiers.TCRdistClassifier import TCRdistClassifier
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.ml_reports.MLReport import MLReport
from immuneML.util.PathBuilder import PathBuilder


class TCRdistMotifDiscovery(MLReport):
    """
    The report for discovering motifs in paired immune receptor data of given specificity based on TCRdist3. The receptors are hierarchically
    clustered based on the tcrdist distance and then motifs are discovered for each cluster. The report outputs logo plots for the motifs along with
    the raw data used for plotting in csv format.

    For the implementation, `TCRdist3 <https://tcrdist3.readthedocs.io/en/latest/>`_ library was used (source code available
    `here <https://github.com/kmayerb/tcrdist3>`_). More details on the functionality used for this report are available
    `here <https://tcrdist3.readthedocs.io/en/latest/motif_gallery.html>`_.

    Original publications:

    Dash P, Fiore-Gartland AJ, Hertz T, et al. Quantifiable predictive features define epitope-specific T cell receptor repertoires.
    Nature. 2017; 547(7661):89-93. `doi:10.1038/nature22383 <https://www.nature.com/articles/nature22383>`_

    Mayer-Blackwell K, Schattgen S, Cohen-Lavi L, et al. TCR meta-clonotypes for biomarker discovery with tcrdist3: quantification of public,
    HLA-restricted TCR biomarkers of SARS-CoV-2 infection. bioRxiv. Published online December 26, 2020:2020.12.24.424260.
    `doi:10.1101/2020.12.24.424260 <https://www.biorxiv.org/content/10.1101/2020.12.24.424260v1>`_


    Specification arguments:

    - positive_class_name (str): the class value (e.g., epitope) used to select only the receptors that are specific to the given epitope so that
      only those sequences are used to infer motifs; the reference receptors as required by TCRdist will be the ones from the dataset that have
      different or no epitope specified in their metadata; if the labels are available only on the epitope level (e.g., label is "AVFDRKSDAK" and
      classes are True and False), then here it should be specified that only the receptors with value "True" for label "AVFDRKSDAK" should be used;
      there is no default value for this argument

    - cores (int): number of processes to use for the computation of the distance and motifs

    - min_cluster_size (int): the minimum size of the cluster to discover the motifs for

    - use_reference_sequences (bool): when showing motifs, this parameter defines if reference sequences should be provided as well as a background


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_tcr_dist_report: # user-defined name
            TCRdistMotifDiscovery:
                positive_class_name: True # class name, could also be epitope name, depending on how it's defined in the dataset
                cores: 4
                min_cluster_size: 30
                use_reference_sequences: False

    """

    @classmethod
    def build_object(cls, **kwargs):
        return TCRdistMotifDiscovery(**kwargs)

    def __init__(self, train_dataset: ReceptorDataset = None, test_dataset: ReceptorDataset = None, method: MLMethod = None, result_path: Path = None,
                 name: str = None, cores: int = None, context: dict = None, positive_class_name=None, min_cluster_size: int = None,
                 use_reference_sequences: bool = None, hp_setting: HPSetting = None, label=None, number_of_processes: int = 1):
        super().__init__(train_dataset=train_dataset, test_dataset=test_dataset, method=method, result_path=result_path,
                         name=name, hp_setting=hp_setting, label=label, number_of_processes=number_of_processes)
        self.cores = cores
        self.positive_class_name = positive_class_name
        self.min_cluster_size = min_cluster_size
        self.use_reference_sequences = use_reference_sequences
        self.context = context

    def _generate(self) -> ReportResult:
        from immuneML.util.TCRdistHelper import TCRdistHelper
        from tcrdist.rep_diff import hcluster_diff
        from tcrdist.summarize import member_summ

        PathBuilder.build(self.result_path)

        subsampled_dataset = self._extract_positive_example_dataset()
        reference_sequences = self._extract_reference_sequences()
        tcr_rep = TCRdistHelper.compute_tcr_dist(subsampled_dataset, [self.label.name], self.cores)
        tcr_rep.hcluster_df, tcr_rep.Z = hcluster_diff(clone_df=tcr_rep.clone_df, pwmat=tcr_rep.pw_alpha + tcr_rep.pw_beta, x_cols=["epitope"],
                                                       count_col='count')

        figures, tables = [], []

        logging.info(f'{TCRdistMotifDiscovery.__name__}: created {tcr_rep.hcluster_df.shape[0]} clusters, now discovering motifs in clusters.')

        for index, row in tcr_rep.hcluster_df.iterrows():
            if len(row['neighbors_i']) >= self.min_cluster_size:
                figure_outputs, table_outputs = self._discover_motif_in_cluster(tcr_rep, index, row, reference_sequences)
                figures.extend(figure_outputs)
                tables.extend(table_outputs)

        res_summary = member_summ(res_df=tcr_rep.hcluster_df, clone_df=tcr_rep.clone_df, addl_cols=['epitope'])
        res_summary.to_csv(self.result_path / "tcrdist_summary.csv")

        tables.append(ReportOutput(path=self.result_path / "tcrdist_summary.csv", name="TCRdist summary (csv)"))

        return ReportResult(name=self.name, info="TCRdist motif discovery", output_figures=figures, output_tables=tables)

    def _discover_motif_in_cluster(self, tcr_rep, index, row, negative_examples=None) -> Tuple[List[ReportOutput], List[ReportOutput]]:
        from tcrdist.adpt_funcs import get_centroid_seq
        from tcrdist.summarize import _select

        from palmotif import compute_pal_motif
        from palmotif import svg_logo

        dfnode = tcr_rep.clone_df.iloc[row['neighbors_i'],]
        figure_outputs, table_outputs = [], []

        logging.info(f"{TCRdistMotifDiscovery.__name__}: in cluster {index+1}, there are {dfnode.shape[0]} neighbors.")

        for chain in ['a', 'b']:

            if dfnode.shape[0] > 2:
                centroid, *_ = get_centroid_seq(df=dfnode)
            else:
                centroid = dfnode[f'cdr3_{chain}_aa'].to_list()[0]

            motif, stat = compute_pal_motif(seqs=_select(df=tcr_rep.clone_df, iloc_rows=row['neighbors_i'], col=f'cdr3_{chain}_aa'),
                                            centroid=centroid, refs=negative_examples[chain] if self.use_reference_sequences else None)

            figure_path = self.result_path / f"motif_{chain}_{index + 1}.svg"
            svg_logo(motif, filename=figure_path)

            motif_data_path = self.result_path / f"motif_{chain}_{index + 1}.csv"
            motif.to_csv(motif_data_path)

            figure_outputs.append(ReportOutput(figure_path, f'Motif {index + 1} ({Chain.get_chain(chain.upper()).name.lower()} chain)'))
            table_outputs.append(ReportOutput(motif_data_path, f'motif {index + 1} ({Chain.get_chain(chain.upper()).name.lower()} chain) csv data'))

        return figure_outputs, table_outputs

    def set_context(self, context: dict):
        self.context = context
        return self

    def check_prerequisites(self):
        if isinstance(self.train_dataset, ReceptorDataset) and isinstance(self.method, TCRdistClassifier):
            return True
        else:
            return False

    def _extract_positive_example_dataset(self) -> ReceptorDataset:
        positive_example_indices = []
        for index, receptor in enumerate(self.train_dataset.get_data()):
            if str(receptor.metadata[self.label.name]) == str(self.positive_class_name):
                positive_example_indices.append(index)

        subsampled_dataset = self.train_dataset.make_subset(example_indices=positive_example_indices, path=self.result_path,
                                                            dataset_type=ReceptorDataset.SUBSAMPLED)

        logging.info(f"{TCRdistMotifDiscovery.__name__}: extracted only positive examples from the training dataset (examples with class = "
                     f"{self.positive_class_name}) for motif discovery. Example count in the new dataset: {subsampled_dataset.get_example_count()}.")

        return subsampled_dataset

    def _extract_reference_sequences(self) -> dict:

        reference_sequences = {'a': [], 'b': []}

        if self.use_reference_sequences:
            for index, receptor in enumerate(self.train_dataset.get_data()):
                if str(receptor.metadata[self.label.name]) != str(self.positive_class_name):
                    reference_sequences['a'].append(receptor.alpha.sequence_aa)
                    reference_sequences['b'].append(receptor.beta.sequence_aa)

        return reference_sequences



import copy
import json
import shutil
import zipfile
from pathlib import Path

import airr
import pandas as pd

from immuneML.IO.dataset_import.AIRRImport import AIRRImport
from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.util.ImportHelper import ImportHelper
from immuneML.util.PathBuilder import PathBuilder
from scripts.specification_util import update_docs_per_mapping


class IReceptorImport(DataImport):
    """
    Imports AIRR datasets retrieved through the `iReceptor Gateway <https://gateway.ireceptor.org/home>`_ into a Repertoire-, Sequence- or ReceptorDataset.
    The differences between this importer and the :ref:`AIRR` importer are:

    * This importer takes in a list of .zip files, which must contain one or more AIRR tsv files, and for each AIRR file, a corresponding metadata json file must be present.
    * This importer does not require a metadata csv file for RepertoireDataset import, it is generated automatically from the metadata json files.

    RepertoireDatasets should be used when making predictions per repertoire, such as predicting a disease state.
    SequenceDatasets or ReceptorDatasets should be used when predicting values for unpaired (single-chain) and paired
    immune receptors respectively, like antigen specificity.

    AIRR rearrangement schema can be found here: https://docs.airr-community.org/en/stable/datarep/rearrangements.html

    When importing a ReceptorDataset, the AIRR field cell_id is used to determine the chain pairs.


    Specification arguments:

    - path (str): This is the path to a directory **with .zip files** retrieved from the iReceptor Gateway. These .zip files should include AIRR files (with .tsv extension) and corresponding metadata.json files with matching names (e.g., for my_dataset.tsv the corresponding metadata file is called my_dataset-metadata.json). The zip files must use the .zip extension.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset or ReceptorDataset. By default, is_repertoire is set to True.

    - paired (str): Required for Sequence- or ReceptorDatasets. This parameter determines whether to import a SequenceDataset (paired = False) or a ReceptorDataset (paired = True). In a ReceptorDataset, two sequences with chain types specified by receptor_chains are paired together based on the identifier given in the AIRR column named 'cell_id'.

    - receptor_chains (str): Required for ReceptorDatasets. Determines which pair of chains to import for each Receptor. Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum. If receptor_chains is not provided, the chain pair is automatically detected (only one chain pair type allowed per repertoire).

    - import_productive (bool): Whether productive sequences (with value 'T' in column productive) should be included in the imported sequences. By default, import_productive is True.

    - import_with_stop_codon (bool): Whether sequences with stop codons (with value 'T' in column stop_codon) should be included in the imported sequences. This only applies if column stop_codon is present. By default, import_with_stop_codon is False.

    - import_out_of_frame (bool): Whether out of frame sequences (with value 'F' in column vj_in_frame) should be included in the imported sequences. This only applies if column vj_in_frame is present. By default, import_out_of_frame is False.

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - region_type (str): Which part of the sequence to import. By default, this value is set to IMGT_CDR3. This means the first and last amino acids are removed from the CDR3 sequence, as AIRR uses the IMGT junction. Specifying any other value will result in importing the sequences as they are. Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - column_mapping (dict): A mapping from AIRR column names to immuneML's internal data representation. For AIRR, this is by default set to the values shown in YAML below.  A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the AIRR file, or using alternative column names). Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the AIRR file, or using alternative column names). Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS.

        .. indent with spaces
        .. code-block:: yaml

                junction: sequences
                junction_aa: sequence_aas
                v_call: v_alleles
                j_call: j_alleles
                locus: chains
                duplicate_count: counts
                sequence_id: sequence_identifiers

    - column_mapping_synonyms (dict): This is a column mapping that can be used if a column could have alternative names. The formatting is the same as column_mapping. If some columns specified in column_mapping are not found in the file, the columns specified in column_mapping_synonyms are instead attempted to be loaded. For AIRR format, there is no default column_mapping_synonyms.

    - metadata_column_mapping (dict): Specifies metadata for Sequence- and ReceptorDatasets. This should specify a mapping similar to column_mapping where keys are AIRR column names and values are the names that are internally used in immuneML as metadata fields. These metadata fields can be used as prediction labels for Sequence- and ReceptorDatasets. For AIRR format, there is no default metadata_column_mapping. When importing a RepertoireDataset, the metadata is automatically extracted from the metadata json files.

    - separator (str): Column separator, for AIRR this is by default "\\t".


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_airr_dataset:
            format: IReceptor
            params:
                path: path/to/zipfiles/
                is_repertoire: True # whether to import a RepertoireDataset
                metadata_column_mapping: # metadata column mapping AIRR: immuneML for Sequence- or ReceptorDatasetDataset
                    airr_column_name1: metadata_label1
                    airr_column_name2: metadata_label2
                import_productive: True # whether to include productive sequences in the dataset
                import_with_stop_codon: False # whether to include sequences with stop codon in the dataset
                import_out_of_frame: False # whether to include out of frame sequences in the dataset
                import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                import_empty_nt_sequences: True # keep sequences even if the `sequences` column is empty (provided that other fields are as specified here)
                import_empty_aa_sequences: False # remove all sequences with empty `sequence_aas` column
                # Optional fields with AIRR-specific defaults, only change when different behavior is required:
                separator: "\\t" # column separator
                region_type: IMGT_CDR3 # what part of the sequence to import
                column_mapping: # column mapping AIRR: immuneML
                    junction: sequences
                    junction_aa: sequence_aas
                    v_call: v_alleles
                    j_call: j_alleles
                    locus: chains
                    duplicate_count: counts
                    sequence_id: sequence_identifiers

    """
    REPERTOIRES_FOLDER = "repertoires/"

    @staticmethod
    def import_dataset(params: dict, dataset_name: str) -> Dataset:
        if params["is_repertoire"]:
            dataset = IReceptorImport.import_repertoire_dataset(params, dataset_name)
        else:
            dataset = IReceptorImport.import_sequence_dataset(params, dataset_name)

        return dataset

    @staticmethod
    def import_repertoire_dataset(params: dict, dataset_name: str) -> RepertoireDataset:
        base_result_path = params['result_path'] / "tmp_airr"
        metadata_file_path = base_result_path / "metadata.csv"

        IReceptorImport._create_airr_repertoiredataset(params['path'], base_result_path, metadata_file_path)

        airr_params = copy.deepcopy(params)
        airr_params["path"] = base_result_path
        airr_params["metadata_file"] = metadata_file_path

        dataset = ImportHelper.import_dataset(AIRRImport, airr_params, dataset_name)

        shutil.rmtree(base_result_path)

        return dataset

    @staticmethod
    def import_sequence_dataset(params: dict, dataset_name: str) -> RepertoireDataset:
        base_result_path = params['result_path'] / "tmp_airr"

        unzipped_path = base_result_path / "tmp_unzipped"
        IReceptorImport._unzip_files(params['path'], unzipped_path, unzip_metadata=False)

        airr_params = copy.deepcopy(params)
        airr_params["path"] = unzipped_path

        dataset = ImportHelper.import_dataset(AIRRImport, airr_params, dataset_name)

        shutil.rmtree(unzipped_path)

        return dataset

    @staticmethod
    def _create_airr_repertoiredataset(input_zips_path: Path, base_result_path: Path, metadata_file_path: Path):
        unzipped_path = base_result_path / "tmp_unzipped/"
        PathBuilder.build(base_result_path / IReceptorImport.REPERTOIRES_FOLDER)

        IReceptorImport._unzip_files(input_zips_path, unzipped_path)

        all_metadata_dfs = []

        for airr_filename in unzipped_path.glob("*.tsv"):
            metadata_filename = unzipped_path / f"{airr_filename.stem}-metadata.json"

            sub_metadata_df = IReceptorImport._create_metadata_df(metadata_filename)
            files_written = IReceptorImport._split_airr_files(airr_filename, sub_metadata_df, base_result_path)
            sub_metadata_df = sub_metadata_df[files_written]

            all_metadata_dfs.append(sub_metadata_df)

        metadata_df = pd.concat(all_metadata_dfs, join="outer", ignore_index=True)
        metadata_df.fillna("NA", inplace=True)
        metadata_df.to_csv(metadata_file_path, index=False)

        shutil.rmtree(unzipped_path)


    @staticmethod
    def _unzip_files(path: Path, unzipped_path: Path, unzip_metadata=True) -> Dataset:
        for zip_filename in path.glob("*.zip"):
            with zipfile.ZipFile(zip_filename, "r") as zip_object:
                for file in zip_object.filelist:
                    file.filename = f"{zip_filename.stem}_{file.filename}"
                    if file.filename.endswith(".tsv") or (file.filename.endswith("-metadata.json") and unzip_metadata):
                        zip_object.extract(file, path=unzipped_path)

    @staticmethod
    def _safe_get_field(dict, nested_fields):
        try:
            result = dict
            for field_name in nested_fields:
                result = result[field_name]
        except KeyError:
            result = None

        return result

    @staticmethod
    def _get_metadata_row(repertoire, sample, data_processing):
        repertoire_id = repertoire['repertoire_id']
        sample_processing_id = sample['sample_processing_id']
        data_processing_id = data_processing['data_processing_id']
        filename = f"{IReceptorImport.REPERTOIRES_FOLDER}{repertoire_id}_{sample_processing_id}_{data_processing_id}.tsv".replace(" ", "-")
        subject_id = repertoire["subject"]["subject_id"]

        study_id = IReceptorImport._safe_get_field(repertoire, ["study", "study_id"])
        species_label = IReceptorImport._safe_get_field(repertoire, ["subject", "species", "label"])
        organism_label = IReceptorImport._safe_get_field(repertoire, ["subject", "organism", "label"])
        sex = IReceptorImport._safe_get_field(repertoire, ["subject", "sex"])
        age_min = IReceptorImport._safe_get_field(repertoire, ["subject", "age_min"])
        age_max = IReceptorImport._safe_get_field(repertoire, ["subject", "age_max"])
        age_event = IReceptorImport._safe_get_field(repertoire, ["subject", "age_event"])
        ancestry_population = IReceptorImport._safe_get_field(repertoire, ["subject", "ancestry_population"])
        ethnicity = IReceptorImport._safe_get_field(repertoire, ["subject", "ethnicity"])
        race = IReceptorImport._safe_get_field(repertoire, ["subject", "race"])
        strain_name = IReceptorImport._safe_get_field(repertoire, ["subject", "strain_name"])

        tissue_label = IReceptorImport._safe_get_field(sample, ["tissue", "label"])
        disease_state_sample = IReceptorImport._safe_get_field(sample, ["disease_state_sample"])
        collection_time_point_relative = IReceptorImport._safe_get_field(sample, ["collection_time_point_relative"])
        collection_time_point_reference = IReceptorImport._safe_get_field(sample, ["collection_time_point_reference"])

        return (filename, subject_id, repertoire_id, sample_processing_id, data_processing_id, study_id, species_label,
                organism_label, sex, age_min, age_max, age_event, ancestry_population, ethnicity,
                race, strain_name, tissue_label, disease_state_sample, collection_time_point_relative,
                collection_time_point_reference)

    @staticmethod
    def _get_static_metadata_df(metadata_dict):
        identifiers = [IReceptorImport._get_metadata_row(repertoire, sample, data_processing)
                       for repertoire in metadata_dict["Repertoire"]
                       for sample in repertoire['sample']
                       for data_processing in repertoire['data_processing']]

        metadata_df = pd.DataFrame(identifiers,
                                   columns=["filename", "subject_id", "repertoire_id", "sample_processing_id",
                                            "data_processing_id", "study_id",
                                            "species_label", "organism_label", "sex", "age_min", "age_max", "age_event",
                                            "ancestry_population", "ethnicity", "race", "strain_name", "tissue_label",
                                            "disease_state_sample", "collection_time_point_relative",
                                            "collection_time_point_reference"])

        metadata_df.dropna(axis=1, how="all", inplace=True)

        return metadata_df

    @staticmethod
    def _add_diagnosis_columns(metadata_df, metadata_dict):
        unique_diseases = set(
            [str(diagnosis["disease_diagnosis"]["label"]) for repertoire in metadata_dict["Repertoire"] for diagnosis in
             repertoire['subject']['diagnosis']])

        id_sorted_repertoires = {repertoire["repertoire_id"]: repertoire for repertoire in metadata_dict["Repertoire"]}

        for disease_diagnosis_label in unique_diseases:
            corrected_label = disease_diagnosis_label.replace(" ", "_")
            metadata_df[corrected_label] = "NA"
            metadata_df[f"{corrected_label}_length"] = None
            metadata_df[f"{corrected_label}_stage"] = None
            metadata_df[f"{corrected_label}_immunogen"] = None

            for repertoire_id in metadata_df["repertoire_id"].unique():
                label_sorted_diagnoses = {str(diagnosis["disease_diagnosis"]["label"]): diagnosis for diagnosis in
                                            id_sorted_repertoires[repertoire_id]["subject"]["diagnosis"]}

                for current_diagnosis_label in label_sorted_diagnoses.keys():
                    if current_diagnosis_label == disease_diagnosis_label:
                        metadata_df.loc[metadata_df["repertoire_id"] == repertoire_id, corrected_label] = \
                        IReceptorImport._safe_get_field(label_sorted_diagnoses, [current_diagnosis_label, "study_group_description"])

                        metadata_df.loc[metadata_df["repertoire_id"] == repertoire_id, f"{corrected_label}_length"] = \
                        IReceptorImport._safe_get_field(label_sorted_diagnoses, [current_diagnosis_label, "disease_length"])

                        metadata_df.loc[metadata_df["repertoire_id"] == repertoire_id, f"{corrected_label}_stage"] = \
                        IReceptorImport._safe_get_field(label_sorted_diagnoses, [current_diagnosis_label, "disease_stage"])

                        metadata_df.loc[metadata_df["repertoire_id"] == repertoire_id, f"{corrected_label}_immunogen"] = \
                        IReceptorImport._safe_get_field(label_sorted_diagnoses, [current_diagnosis_label, "immunogen"])

        metadata_df.dropna(axis=1, how="all", inplace=True)

        return metadata_df

    @staticmethod
    def _create_metadata_df(metadata_json):
        with open(metadata_json) as json_file:
            metadata_dict = json.load(json_file)

        metadata_df = IReceptorImport._get_static_metadata_df(metadata_dict)
        metadata_df = IReceptorImport._add_diagnosis_columns(metadata_df, metadata_dict)

        return metadata_df

    @staticmethod
    def _split_airr_files(airr_file: Path, metadata_df: pd.DataFrame, result_path: Path):
        airr_df = airr.load_rearrangement(airr_file)
        files_written = []

        for filename, repertoire_id, sample_processing_id, data_processing_id in metadata_df[
            ["filename", "repertoire_id", "sample_processing_id", "data_processing_id"]].itertuples(index=False):

            subset = airr_df[airr_df["repertoire_id"] == repertoire_id]

            if "sample_processing_id" in subset.columns and any(subset["sample_processing_id"].str.len() > 0):
                subset = subset[subset["sample_processing_id"] == str(sample_processing_id)]
            if "data_processing_id" in subset.columns and any(subset["data_processing_id"].str.len() > 0):
                subset = subset[subset["data_processing_id"] == str(data_processing_id)]

            if subset.empty:
                files_written.append(False)
            else:
                subset.to_csv(result_path / filename, index=False, sep="\t")
                files_written.append(True)

        return files_written


    @staticmethod
    def get_documentation():
        doc = str(IReceptorImport.__doc__)

        chain_pair_values = str([chain_pair.name for chain_pair in ChainPair])[1:-1].replace("'", "`")
        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")
        repertoire_fields = list(Repertoire.FIELDS)
        repertoire_fields.remove("region_type")

        mapping = {
            "Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum.": f"Valid values are {chain_pair_values}.",
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
            "Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS": f"Valid immuneML fields that can be specified here are {repertoire_fields}."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc



import itertools
import warnings
from typing import List

from immuneML.data_model.receptor.BCKReceptor import BCKReceptor
from immuneML.data_model.receptor.BCReceptor import BCReceptor
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.data_model.receptor.Receptor import Receptor
from immuneML.data_model.receptor.TCABReceptor import TCABReceptor
from immuneML.data_model.receptor.TCGDReceptor import TCGDReceptor
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence


class ReceptorBuilder:

    @classmethod
    def build_object(cls, sequences: dict, identifier: str = None, metadata: dict = None) -> Receptor:
        if all(chain in ChainPair.TRA_TRB.value for chain in sequences.keys()):
            return TCABReceptor(alpha=sequences[Chain.ALPHA.value], beta=sequences[Chain.BETA.value], identifier=identifier, metadata=metadata)
        elif all(chain in ChainPair.TRG_TRD.value for chain in sequences.keys()):
            return TCGDReceptor(gamma=sequences[Chain.GAMMA.value], delta=sequences[Chain.DELTA.value], identifier=identifier, metadata=metadata)
        elif all(chain in ChainPair.IGH_IGL.value for chain in sequences.keys()):
            return BCReceptor(heavy=sequences[Chain.HEAVY.value], light=sequences[Chain.LIGHT.value], identifier=identifier, metadata=metadata)
        elif all(chain in ChainPair.IGH_IGK.value for chain in sequences.keys()):
            return BCKReceptor(heavy=sequences[Chain.HEAVY.value], kappa=sequences[Chain.KAPPA.value], identifier=identifier, metadata=metadata)
        else:
            warnings.warn(f"ReceptorBuilder: attempt to build_from_objects receptor with chains {sequences.keys()}, returning None...")
            return None

    @classmethod
    def build_objects(cls, sequences: List[ReceptorSequence]) -> List[Receptor]:
        receptors = []
        sequences_per_chain = {chain.value: [sequence for sequence in sequences if sequence.metadata.chain.value == chain.value]
                               for chain in Chain}
        for chain_pair in ChainPair:
            all_chain_1 = sequences_per_chain[chain_pair.value[0]]
            all_chain_2 = sequences_per_chain[chain_pair.value[1]]
            combinations = list(itertools.product(all_chain_1, all_chain_2))
            receptors.extend([ReceptorBuilder.build_object({sequence.metadata.chain.value: sequence for sequence in combination})
                              for combination in combinations])

        return receptors

    @classmethod
    def build_objects_from_pairs(cls, sequences1: List[ReceptorSequence], sequences2: List[ReceptorSequence]) -> List[Receptor]:
        receptors = [ReceptorBuilder.build_object({sequences1[ind].metadata.chain.value: sequences1[ind],
                                                   sequences2[ind].metadata.chain.value: sequences2[ind]},
                                                  metadata={**sequences1[ind].metadata.custom_params, **sequences2[ind].metadata.custom_params})
                     for ind in range(len(sequences1))]

        return receptors



import logging
import os
import warnings
import zipfile
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pandas as pd
import yaml

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class DesignMatrixExporter(EncodingReport):
    """
    Exports the design matrix and related information of a given encoded Dataset to csv files.
    If the encoded data has more than 2 dimensions (such as when using the OneHot encoder with option Flatten=False),
    the data are then exported to different formats to facilitate their import with external software.

    Specification arguments:

    - file_format (str): the format and extension of the file to store the design matrix. The supported formats are:
      npy, csv, hdf5, npy.zip, csv.zip or hdf5.zip.

    Note: when using hdf5 or hdf5.zip output formats, make sure the 'hdf5' dependency is installed.


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_dme_report:
            DesignMatrixExporter:
                file_format: csv

    """
    def __init__(self, dataset: Dataset = None, result_path: Path = None, file_format: str = None, number_of_processes: int = 1, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.file_format = file_format

    @classmethod
    def build_object(cls, **kwargs):
        ParameterValidator.assert_keys_present(list(kwargs.keys()), ['file_format', 'name'], DesignMatrixExporter.__name__, DesignMatrixExporter.__name__)
        ParameterValidator.assert_in_valid_list(kwargs['file_format'], ['npy', 'csv', 'npy.zip', 'csv.zip', 'hdf5.zip'], DesignMatrixExporter.__name__, 'file_format')

        return DesignMatrixExporter(**kwargs)

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)

        matrix_result = self._export_matrix()
        details_result = self._export_details()
        label_result = self._export_labels()

        return ReportResult(self.name,
                            info="The design matrix and related information of a given encoded Dataset",
                            output_tables=[matrix_result, label_result], output_text=[details_result])

    def _export_matrix(self) -> ReportOutput:
        """Create a file for the design matrix in the desired format."""
        
        data = self._get_data()
        file_path = self.result_path / "design_matrix"
        ext = os.path.splitext(self.file_format)[0]
        file_path = file_path.with_suffix('.' + ext)

        # Use h5py to create a hdf5 file.
        if ext == "hdf5":
            import h5py
            with h5py.File(str(file_path), 'w') as hf_object:
                hf_object.create_dataset(str(file_path), data=data)
        # Use numpy to create a csv or npy file.
        elif len(data.shape) <= 2 and ext == "csv":
            feature_names = self.dataset.encoded_data.feature_names
            header = ",".join(str(name) for name in feature_names) if feature_names is not None else ""
            np.savetxt(fname=str(file_path), X=data, delimiter=",", comments='',
                       header=header)
        else:
            if ext != "npy":
                logging.info('The selected Report format is not compatible, .npy is used instead')
                file_path = file_path.with_suffix(".npy")
                ext = "npy"
            np.save(str(file_path), data)
        
        # If requested, compress the file into a .zip.
        if self.file_format.endswith(".zip"):
            file_path_zip = file_path.with_suffix('.' + ext + '.zip')
            with zipfile.ZipFile(str(file_path_zip), 'w') as zipped_file:
                zipped_file.write(str(file_path), compress_type=zipfile.ZIP_DEFLATED)
            os.remove(str(file_path)) 
            file_path = file_path_zip
        return ReportOutput(file_path, "design matrix")

    def _get_data(self) -> np.ndarray:
        if isinstance(self.dataset.encoded_data.examples, np.ndarray):
            data = self.dataset.encoded_data.examples
        elif isinstance(self.dataset.encoded_data.examples, pd.DataFrame):
            data = self.dataset.encoded_data.examples.to_numpy()
        else: #scipy
            data = self.dataset.encoded_data.examples.toarray()
        return data

    def _export_details(self) -> ReportOutput:
        file_path = self.result_path / "encoding_details.yaml"
        with file_path.open("w") as file:
            details = {
                "feature_names": self.dataset.encoded_data.feature_names,
                "encoding": self.dataset.encoded_data.encoding,
                "example_ids": list(self.dataset.encoded_data.example_ids)
            }

            yaml.dump(details, file)

        return ReportOutput(file_path, "encoding details")

    def _export_labels(self) -> ReportOutput:
        if self.dataset.encoded_data.labels is not None:
            labels_df = pd.DataFrame(self.dataset.encoded_data.labels)
            file_path = self.result_path / "labels.csv"
            labels_df.to_csv(file_path, sep=",", index=False)
            return ReportOutput(file_path, "exported labels")

    def check_prerequisites(self):
        if self.dataset.encoded_data is None or self.dataset.encoded_data.examples is None:
            warnings.warn("DesignMatrixExporter: the dataset is not encoded, skipping this report...")
            return False
        else:
            return True



import copy
import inspect
from pathlib import Path
from typing import List

import sklearn.metrics as sklearn_metrics

from immuneML.dsl.instruction_parsers.LabelHelper import LabelHelper
from immuneML.dsl.symbol_table.SymbolTable import SymbolTable
from immuneML.dsl.symbol_table.SymbolType import SymbolType
from immuneML.environment.LabelConfiguration import LabelConfiguration
from immuneML.ml_methods.clustering.ClusteringMethod import ClusteringMethod
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod
from immuneML.reports.Report import Report
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.workflows.instructions.clustering.ClusteringInstruction import ClusteringInstruction
from immuneML.workflows.instructions.clustering.clustering_run_model import ClusteringSetting


class ClusteringParser:

    def parse(self, key: str, instruction: dict, symbol_table: SymbolTable, path: Path = None) -> ClusteringInstruction:
        valid_keys = [k for k in inspect.signature(ClusteringInstruction.__init__).parameters.keys()
                      if k not in ['result_path', 'name', 'self', 'label_config']] + ['type', 'labels']

        ParameterValidator.assert_keys(instruction.keys(), valid_keys, ClusteringParser.__name__, key)

        dataset = symbol_table.get(instruction['dataset'])
        clustering_settings = parse_clustering_settings(key, instruction, symbol_table)
        metrics = parse_metrics(key, instruction, symbol_table)
        label_config = parse_labels(key, instruction, dataset)
        reports = parse_reports(key, instruction, symbol_table)

        return ClusteringInstruction(dataset, metrics, clustering_settings, key, label_config, reports)


def parse_labels(key, instruction, dataset) -> LabelConfiguration:
    if 'labels' in instruction and instruction['labels'] is not None:
        label_config = LabelHelper.create_label_config(instruction['labels'], dataset, key, 'labels')
        return label_config


def parse_reports(key, instruction, symbol_table) -> List[Report]:
    if 'reports' not in instruction or instruction['reports'] is None:
        return []
    else:
        ParameterValidator.assert_type_and_value(instruction['reports'], list, 'ClusteringParser', 'reports')
        valid_reports = symbol_table.get_keys_by_type(SymbolType.REPORT)

        ParameterValidator.assert_all_type_and_value(instruction['reports'], str, 'ClusteringParser', 'reports')
        ParameterValidator.assert_all_in_valid_list(instruction['reports'], valid_reports, 'ClusteringParser', 'reports')

        reports = [symbol_table.get(report_id) for report_id in instruction['reports']]
        return reports


def parse_metrics(key: str, instruction: dict, symbol_table: SymbolTable) -> List[str]:
    ParameterValidator.assert_type_and_value(instruction['metrics'], list, 'ClusteringParser', f'{key}:metrics')
    ParameterValidator.assert_all_type_and_value(instruction['metrics'], str, 'ClusteringParser', f'{key}:metrics')

    for metric in instruction['metrics']:
        assert hasattr(sklearn_metrics, metric), (f"Clustering parser: metric {metric} is not a valid metric. See the "
                                                  f"list of scikit-learn's metrics for clustering.")

    return instruction['metrics']


def parse_clustering_settings(key: str, instruction: dict, symbol_table: SymbolTable) -> List[ClusteringSetting]:
    ParameterValidator.assert_type_and_value(instruction['clustering_settings'], list, 'ClusteringParser', 'key:clustering_settings')
    valid_encodings = symbol_table.get_keys_by_type(SymbolType.ENCODING)
    valid_dim_red = [method.symbol for method in symbol_table.get_by_type(SymbolType.ML_METHOD)
                     if isinstance(method.item, DimRedMethod)]
    valid_clusterings = [method.symbol for method in symbol_table.get_by_type(SymbolType.ML_METHOD)
                         if isinstance(method.item, ClusteringMethod)]

    settings_objs = []
    for setting in instruction['clustering_settings']:
        setting_obj = make_setting_obj(setting, valid_encodings, valid_clusterings, valid_dim_red, symbol_table,
                                       instruction)
        settings_objs.append(setting_obj)

    return settings_objs


def make_setting_obj(setting, valid_encodings, valid_clusterings, valid_dim_red, symbol_table, instruction):
    ParameterValidator.assert_keys(setting.keys(), ['encoding', 'dim_reduction', 'method'], 'ClusteringParser', 'clustering_settings')

    ParameterValidator.assert_in_valid_list(setting['encoding'], valid_encodings, 'ClusteringParser', 'encoding')
    ParameterValidator.assert_in_valid_list(setting['method'], valid_clusterings, 'ClusteringParser',
                                            'method')
    if 'dim_reduction' in setting and setting['dim_reduction'] is not None:
        ParameterValidator.assert_in_valid_list(setting['dim_reduction'], valid_dim_red, 'ClusteringParser',
                                                'dim_reduction')
        dim_reduction = copy.deepcopy(symbol_table.get(setting['dim_reduction']))
        dim_red_params = symbol_table.get_config(setting['dim_reduction'])
        dim_red_name = setting['dim_reduction']
    else:
        dim_reduction, dim_red_params, dim_red_name = None, None, None

    encoder = make_encoder_obj(symbol_table, setting['encoding'], instruction['dataset'])
    method = copy.deepcopy(symbol_table.get(setting['method']))

    return ClusteringSetting(encoder=encoder, encoder_params=symbol_table.get_config(setting['encoding']),
                             encoder_name=setting['encoding'], clustering_method=method,
                             clustering_params=symbol_table.get_config(setting['method']),
                             clustering_method_name=setting['method'], dim_reduction_method=dim_reduction,
                             dim_red_params=dim_red_params, dim_red_name=dim_red_name)


def make_encoder_obj(symbol_table, encoding_key, dataset_key):
    return symbol_table.get(encoding_key).build_object(symbol_table.get(dataset_key),
                                                       **symbol_table.get_config(encoding_key)["encoder_params"]) \
        .set_context({"dataset": symbol_table.get(dataset_key)})



import warnings
from collections import Counter
from pathlib import Path

import pandas as pd
import plotly.express as px
import numpy as np

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.ReceptorDataset import ReceptorDataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.dataset.SequenceDataset import SequenceDataset
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence
from immuneML.environment.EnvironmentSettings import EnvironmentSettings
from immuneML.environment.SequenceType import SequenceType
from immuneML.reports.PlotlyUtil import PlotlyUtil
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder
from immuneML.util.PositionHelper import PositionHelper


class AminoAcidFrequencyDistribution(DataReport):
    """
    Generates a barplot showing the relative frequency of each amino acid at each position in the sequences of a dataset.

    Specification arguments:

    - imgt_positions (bool): Whether to use IMGT positional numbering or sequence index numbering. When imgt_positions
      is True, IMGT positions are used, meaning sequences of unequal length are aligned according to their IMGT
      positions. By default, imgt_positions is True.

    - relative_frequency (bool): Whether to plot relative frequencies (true) or absolute counts (false) of the
      positional amino acids. By default, relative_frequency is True.

    - split_by_label (bool): Whether to split the plots by a label. If set to true, the Dataset must either contain a
      single label, or alternatively the label of interest can be specified under 'label'. If split_by_label is set to
      true, the percentage-wise frequency difference between classes is plotted additionally. By default,
      split_by_label is False.

    - label (str): if split_by_label is set to True, a label can be specified here.

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_aa_freq_report:
          AminoAcidFrequencyDistribution:
            relative_frequency: False
            split_by_label: True
            label: CMV

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = AminoAcidFrequencyDistribution.__name__
        ParameterValidator.assert_type_and_value(kwargs["imgt_positions"], bool, location, "imgt_positions")
        ParameterValidator.assert_type_and_value(kwargs["relative_frequency"], bool, location, "relative_frequency")
        ParameterValidator.assert_type_and_value(kwargs["split_by_label"], bool, location, "split_by_label")

        if kwargs["label"] is not None:
            ParameterValidator.assert_type_and_value(kwargs["label"], str, location, "label")

            if kwargs["split_by_label"] is False:
                warnings.warn(f"{location}: label is set but split_by_label was False, setting split_by_label to True")
                kwargs["split_by_label"] = True

        return AminoAcidFrequencyDistribution(**kwargs)

    def __init__(self, dataset: Dataset = None, imgt_positions: bool = None, relative_frequency: bool = None,
                 split_by_label: bool = None, label: str = None,
                 result_path: Path = None, number_of_processes: int = 1, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.imgt_positions = imgt_positions
        self.relative_frequency = relative_frequency
        self.split_by_label = split_by_label
        self.label_name = label

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)

        freq_dist = self._get_plotting_data()

        tables = []
        figures = []

        tables.append(self._write_output_table(freq_dist,
                                               self.result_path / "amino_acid_frequency_distribution.tsv",
                                               name="Table of amino acid frequencies"))

        figures.append(self._safe_plot(freq_dist=freq_dist, plot_callable="_plot_distribution"))

        if self.split_by_label:
            frequency_change = self._compute_frequency_change(freq_dist)

            tables.append(self._write_output_table(frequency_change,
                                                   self.result_path / f"frequency_change.tsv",
                                                   name=f"Frequency change between classes"))
            figures.append(self._safe_plot(frequency_change=frequency_change, plot_callable="_plot_frequency_change"))

        return ReportResult(name=self.name,
                            info="A barplot showing the relative frequency of each amino acid at each position in the sequences of a dataset.",
                            output_figures=[fig for fig in figures if fig is not None],
                            output_tables=[table for table in tables if table is not None])

    def _get_plotting_data(self):
        if isinstance(self.dataset, SequenceDataset):
            plotting_data = self._get_sequence_dataset_plotting_data()

        elif isinstance(self.dataset, ReceptorDataset):
            plotting_data = self._get_receptor_dataset_plotting_data()

        elif isinstance(self.dataset, RepertoireDataset):
            plotting_data = self._get_repertoire_dataset_plotting_data()

        if not self.split_by_label:
            plotting_data.drop(columns=["class"], inplace=True)

        return plotting_data

    def _get_sequence_dataset_plotting_data(self):
        return self._count_dict_per_class_to_df(self._count_aa_frequencies(self._sequence_class_iterator()))

    def _count_dict_per_class_to_df(self, raw_count_dict_per_class):
        result_dfs = []

        for class_name, raw_count_dict in raw_count_dict_per_class.items():
            result_df = self._count_dict_to_df(raw_count_dict)
            result_df["class"] = class_name
            result_dfs.append(result_df)

        return pd.concat(result_dfs)

    def _sequence_class_iterator(self):
        label_name = self._get_label_name()

        for sequence in self.dataset.get_data():
            if self.split_by_label:
                yield (sequence, sequence.get_attribute(label_name))
            else:
                yield (sequence, 0)

    def _get_receptor_dataset_plotting_data(self):
        result_dfs = []

        receptors = self.dataset.get_data()
        chains = next(receptors).get_chains()

        for chain in chains:
            raw_count_dict_per_class = self._count_aa_frequencies(self._chain_class_iterator(chain))

            for class_name, raw_count_dict in raw_count_dict_per_class.items():
                result_df = self._count_dict_to_df(raw_count_dict)
                result_df["chain"] = chain
                result_df["class"] = class_name
                result_dfs.append(result_df)

        return pd.concat(result_dfs)

    def _chain_class_iterator(self, chain):
        label_name = self._get_label_name()

        for receptor in self.dataset.get_data():
            assert chain in receptor.get_chains(), f"{AminoAcidFrequencyDistribution.__name__}: All receptors in the dataset must contain the same chains. Expected {chain} but found {receptor.get_chains()}"

            if self.split_by_label:
                yield (receptor.get_chain(chain), receptor.get_attribute(label_name))
            else:
                yield (receptor.get_chain(chain), 0)

    def _get_repertoire_dataset_plotting_data(self):
        raw_count_dict_per_class = {}
        label_name = self._get_label_name()

        if self.split_by_label:
            class_names = self.dataset.get_metadata([label_name])[label_name]
        else:
            class_names = [0] * self.dataset.get_example_count()

        for repertoire, class_name in zip(self.dataset.get_data(), class_names):
            self._count_aa_frequencies(self._repertoire_class_iterator(repertoire, class_name),
                                       raw_count_dict_per_class)

        return self._count_dict_per_class_to_df(raw_count_dict_per_class)

    def _repertoire_class_iterator(self, repertoire, class_name):
        for sequence in repertoire.get_sequence_objects():
            yield (sequence, class_name)

    def _count_aa_frequencies(self, sequence_class_iterator, raw_count_dict=None):
        raw_count_dict = {} if raw_count_dict is None else raw_count_dict

        for item in sequence_class_iterator:
            sequence, class_name = item
            seq_str = sequence.get_sequence(sequence_type=SequenceType.AMINO_ACID)
            seq_pos = self._get_positions(sequence)

            if class_name not in raw_count_dict:
                raw_count_dict[class_name] = {}

            for aa, pos in zip(seq_str, seq_pos):
                if pos not in raw_count_dict[class_name]:
                    raw_count_dict[class_name][pos] = {legal_aa: 0 for legal_aa in
                                                       EnvironmentSettings.get_sequence_alphabet(
                                                           SequenceType.AMINO_ACID)}

                raw_count_dict[class_name][pos][aa] += 1

        return raw_count_dict

    def _count_dict_to_df(self, raw_count_dict):
        df_dict = {"amino acid": [], "position": [], "count": [], "relative frequency": []}
        for pos in raw_count_dict:
            df_dict["position"].extend([pos] * 20)
            df_dict["amino acid"].extend(list(raw_count_dict[pos].keys()))

            counts = list(raw_count_dict[pos].values())
            total_count_for_pos = sum(counts)

            df_dict["count"].extend(counts)
            df_dict["relative frequency"].extend([count / total_count_for_pos for count in counts])

        return pd.DataFrame(df_dict)

    def _get_positions(self, sequence: ReceptorSequence):
        if self.imgt_positions:
            positions = PositionHelper.gen_imgt_positions_from_length(
                len(sequence.get_sequence(SequenceType.AMINO_ACID)),
                sequence.get_attribute("region_type"))
        else:
            positions = list(range(1, len(sequence.get_sequence(SequenceType.AMINO_ACID)) + 1))

        return [str(pos) for pos in positions]

    def _write_results_table(self, results_table):
        file_path = self.result_path / "amino_acid_frequency_distribution.tsv"

        results_table.to_csv(file_path, index=False)

        return ReportOutput(path=file_path, name="Table of amino acid frequencies")

    def _plot_distribution(self, freq_dist):
        freq_dist.sort_values(by=["amino acid"], ascending=False, inplace=True)
        category_orders = None if "class" not in freq_dist.columns else {"class": sorted(set(freq_dist["class"]))}

        y = "relative frequency" if self.relative_frequency else "count"

        figure = px.bar(freq_dist, x="position", y=y, color="amino acid", text="amino acid",
                        facet_col="class" if "class" in freq_dist.columns else None,
                        facet_row="chain" if "chain" in freq_dist.columns else None,
                        color_discrete_map=PlotlyUtil.get_amino_acid_color_map(),
                        category_orders=category_orders,
                        labels={"position": "IMGT position" if self.imgt_positions else "Position",
                                "count": "Count",
                                "relative frequency": "Relative frequency",
                                "amino acid": "Amino acid"}, template="plotly_white")
        figure.update_xaxes(categoryorder='array', categoryarray=self._get_position_order(freq_dist["position"]))
        figure.update_layout(showlegend=False, yaxis={'categoryorder': 'category ascending'})

        if self.relative_frequency:
            figure.update_yaxes(tickformat=",.0%", range=[0, 1])

        file_path = self.result_path / "amino_acid_frequency_distribution.html"
        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name="Amino acid frequency distribution")

    def _get_position_order(self, positions):
        return [str(int(pos)) if pos.is_integer() else str(pos) for pos in sorted(set(positions.astype(float)))]

    def _compute_frequency_change(self, freq_dist):
        classes = sorted(set(freq_dist["class"]))
        assert len(classes) == 2, f"{AminoAcidFrequencyDistribution.__name__}: cannot compute frequency change when the number of classes is not 2: {classes}"

        class_a_df = freq_dist[freq_dist["class"] == classes[0]]
        class_b_df = freq_dist[freq_dist["class"] == classes[1]]

        on = ["amino acid", "position"]
        on = on + ["chain"] if "chain" in freq_dist.columns else on

        merged_dfs = pd.merge(class_a_df, class_b_df, on=on, how="outer", suffixes=["_a", "_b"])
        merged_dfs = merged_dfs[(merged_dfs["relative frequency_a"] + merged_dfs["relative frequency_b"]) > 0]

        merged_dfs["frequency_change"] = merged_dfs["relative frequency_a"] - merged_dfs["relative frequency_b"]

        pos_class_a = merged_dfs[merged_dfs["frequency_change"] > 0]
        pos_class_b = merged_dfs[merged_dfs["frequency_change"] < 0]

        pos_class_a["positive_class"] = classes[0]
        pos_class_b["positive_class"] = classes[1]
        pos_class_b["frequency_change"] = 0 - pos_class_b["frequency_change"]

        keep_cols = on + ["frequency_change", "positive_class"]
        pos_class_a = pos_class_a[keep_cols]
        pos_class_b = pos_class_b[keep_cols]

        return pd.concat([pos_class_a, pos_class_b])

    def _plot_frequency_change(self, frequency_change):
        figure = px.bar(frequency_change, x="position", y="frequency_change", color="amino acid", text="amino acid",
                        facet_col="positive_class",
                        facet_row="chain" if "chain" in frequency_change.columns else None,
                        color_discrete_map=PlotlyUtil.get_amino_acid_color_map(),
                        labels={"position": "IMGT position" if self.imgt_positions else "Position",
                                "positive_class": "Class",
                                "frequency_change": "Difference in relative frequency",
                                "amino acid": "Amino acid"}, template="plotly_white")

        figure.update_xaxes(categoryorder='array', categoryarray=self._get_position_order(frequency_change["position"]))
        figure.update_layout(showlegend=False, yaxis={'categoryorder': 'category ascending'})

        figure.update_yaxes(tickformat=",.0%")

        file_path = self.result_path / "frequency_change.html"
        figure.write_html(str(file_path))

        return ReportOutput(path=file_path, name="Frequency difference between amino acid usage in the two classes")

    def _get_label_name(self):
        if self.split_by_label:
            if self.label_name is None:
                return list(self.dataset.get_label_names())[0]
            else:
                return self.label_name
        else:
            return None

    def check_prerequisites(self):
        if self.split_by_label:
            if self.label_name is None:
                if len(self.dataset.get_label_names()) != 1:
                    warnings.warn(f"{AminoAcidFrequencyDistribution.__name__}: ambiguous label: split_by_label was set to True but no label name was specified, and the number of available labels is {len(self.dataset.get_label_names())}: {self.dataset.get_label_names()}. Skipping this report...")
                    return False
            else:
                if self.label_name not in self.dataset.get_label_names():
                    warnings.warn(f"{AminoAcidFrequencyDistribution.__name__}: the specified label name ({self.label_name}) was not available among the dataset labels: {self.dataset.get_label_names()}. Skipping this report...")
                    return False

        return True



import abc
from pathlib import Path

import numpy as np
import pandas as pd

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.encodings.DatasetEncoder import DatasetEncoder
from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.util.EncoderHelper import EncoderHelper
from immuneML.util.ReadsType import ReadsType
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.ReflectionHandler import ReflectionHandler
from scripts.specification_util import update_docs_per_mapping


class MatchedRegexEncoder(DatasetEncoder):
    """
    Encodes the dataset based on the matches between a RepertoireDataset and a collection of regular expressions.
    For each regular expression, the number of sequences in the RepertoireDataset containing the expression is counted.
    This can also be used to count how often a subsequence occurs in a RepertoireDataset.

    The regular expressions are defined per chain, and it is possible to require a V gene match in addition to the
    CDR3 sequence containing the regular expression.

    This encoding can be used in combination with the :ref:`Matches` report.


    Specification arguments:

    - match_v_genes (bool): Whether V gene matches are required. If this is True, a match is only counted if the
      V gene matches the gene specified in the motif input file. By default match_v_genes is False.

    - reads (:py:mod:`~immuneML.util.ReadsType`): Reads type signify whether the counts of the sequences in the
      repertoire will be taken into account. If :py:mod:`~immuneML.util.ReadsType.UNIQUE`, only unique sequences
      (clonotypes) are counted, and if :py:mod:`~immuneML.util.ReadsType.ALL`, the sequence 'count' value is
      summed when determining the number of matches. The default value for reads is all.

    - motif_filepath (str): The path to the motif input file. This should be a tab separated file containing a
      column named 'id' and for every chain that should be matched a column containing the regex (<chain>_regex) and a
      column containing the V gene (<chain>V) if match_v_genes is True.
      The chains are specified by their three-letter code, see :py:obj:`~immuneML.data_model.receptor.receptor_sequence.Chain.Chain`.

    In the simplest case, when counting the number of occurrences of a given list of k-mers in TRB sequences, the
    contents of the motif file could look like this:

        ====  ==========
        id    TRB_regex
        ====  ==========
        1     ACG
        2     EDNA
        3     DFWG
        ====  ==========

    It is also possible to test whether paired regular expressions occur in the dataset (for example: regular expressions
    matching both a TRA chain and a TRB chain) by specifying them on the same line.
    In a more complex case where both paired and unpaired regular expressions are specified, in addition to matching the V
    genes, the contents of the motif file could look like this:

        ====  ==========  =======  ==========  ========
        id    TRA_regex   TRAV     TRB_regex   TRBV
        ====  ==========  =======  ==========  ========
        1     AGQ.GSS     TRAV35   S[APL]GQY   TRBV29-1
        2                          ASS.R.*     TRBV7-3
        ====  ==========  =======  ==========  ========


    YAML Specification:

    .. indent with spaces
    .. code-block:: yaml

        my_mr_encoding:
            MatchedRegex:
                motif_filepath: path/to/file.txt
                match_v_genes: True
                reads: unique

    """

    dataset_mapping = {
        "RepertoireDataset": "MatchedRegexRepertoireEncoder"
    }

    def __init__(self, motif_filepath: Path, match_v_genes: bool, reads: ReadsType, chains: list, name: str = None):
        self.motif_filepath = motif_filepath
        self.match_v_genes = match_v_genes
        self.reads = reads
        self.chains = chains
        self.regex_df = None
        self.feature_count = None
        self.name = name

    @staticmethod
    def _prepare_parameters(motif_filepath: str, match_v_genes: bool, reads: str, name: str = None):

        ParameterValidator.assert_type_and_value(match_v_genes, bool, "MatchedRegexEncoder", "match_v_genes")
        ParameterValidator.assert_in_valid_list(reads.upper(), [item.name for item in ReadsType], "MatchedRegexEncoder", "reads")

        motif_filepath = Path(motif_filepath)
        assert motif_filepath.is_file(), f"MatchedRegexEncoder: the file {motif_filepath} does not exist. " \
                                               f"Specify the correct path under motif_filepath."

        file_columns = list(pd.read_csv(motif_filepath, sep="\t", iterator=False, dtype=str, nrows=0).columns)

        ParameterValidator.assert_all_in_valid_list(file_columns, ["id"] + [f"{c.value}V" for c in Chain] + [f"{c.value}_regex" for c in Chain], "MatchedRegexEncoder", "motif_filepath (column names)")

        chains = [colname.split("_")[0] for colname in file_columns if colname.endswith("_regex")]
        if match_v_genes:
            for chain in chains:
                assert f"{chain}V" in file_columns, f"MatchedRegexEncoder: expected column {chain}V to be present in the columns of motif_filepath. " \
                                                    f"Remove {chain}_regex from columns, or set match_v_genes to False."

        return {
            "motif_filepath": motif_filepath,
            "match_v_genes": match_v_genes,
            "reads": ReadsType[reads.upper()],
            "chains": chains,
            "name": name
        }

    @staticmethod
    def build_object(dataset=None, **params):
        EncoderHelper.check_dataset_type_available_in_mapping(dataset, MatchedRegexEncoder)

        prepared_params = MatchedRegexEncoder._prepare_parameters(**params)
        encoder = ReflectionHandler.get_class_by_name(MatchedRegexEncoder.dataset_mapping[dataset.__class__.__name__], "reference_encoding/")(**prepared_params)

        return encoder

    def encode(self, dataset, params: EncoderParams):
        cache_key = CacheHandler.generate_cache_key(self._prepare_caching_params(dataset, params))
        encoded_dataset = CacheHandler.memo(cache_key,
                                            lambda: self._encode_new_dataset(dataset, params))

        return encoded_dataset

    def _prepare_caching_params(self, dataset, params: EncoderParams):
        return (("dataset_identifiers", tuple(dataset.get_example_ids())),
                ("dataset_metadata", dataset.metadata_file),
                ("dataset_type", dataset.__class__.__name__),
                ("labels", tuple(params.label_config.get_labels_by_name())),
                ("encoding", MatchedRegexEncoder.__name__),
                ("learn_model", params.learn_model),
                ("encoding_params", tuple(vars(self).items())))

    @abc.abstractmethod
    def _encode_new_dataset(self, dataset, params: EncoderParams):
        pass

    def _load_regex_df(self):
        df = pd.read_csv(self.motif_filepath, sep="\t", iterator=False, dtype=str)

        if not self.match_v_genes:
            for v_gene in [f"{c.value}V" for c in Chain]:
                if v_gene in df.columns:
                    df.drop(v_gene, axis=1, inplace=True)

        colnames_subset = list(df.columns)
        colnames_subset.remove("id")
        df.drop_duplicates(subset=colnames_subset, inplace=True)

        df.replace({np.NaN: None}, inplace=True)

        self.feature_count = 0

        for chain in Chain:
            regex_colname = f"{chain.value}_regex"
            if regex_colname in df.columns:
                self.feature_count += df[regex_colname].count()

        self.regex_df = df


    @staticmethod
    def get_documentation():
        doc = str(MatchedRegexEncoder.__doc__)

        chain_values = str([region_type.value for region_type in Chain])[1:-1].replace("'", "`")

        mapping = {
            "The chains are specified by their three letter code, see :py:obj:`~immuneML.data_model.receptor.receptor_sequence.Chain.Chain`.": f"The chains are specified by their three letter code, valid values are: {chain_values}.",
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc






import copy
from dataclasses import fields as get_fields
from multiprocessing.pool import Pool
from pathlib import Path
from uuid import uuid4

import pandas as pd

from immuneML.data_model.bnp_util import write_yaml
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.environment.SequenceType import SequenceType
from immuneML.preprocessing.filters.CountAggregationFunction import CountAggregationFunction
from immuneML.preprocessing.filters.Filter import Filter
from immuneML.util.ParameterValidator import ParameterValidator
from scripts.specification_util import update_docs_per_mapping


class DuplicateSequenceFilter(Filter):
    """
    Collapses duplicate nucleotide or amino acid sequences within each repertoire in the given RepertoireDataset.
    This filter can be applied to Repertoires and RepertoireDatasets.

    Sequences are considered duplicates if the following fields are identical:

    - amino acid or nucleotide sequence (whichever is specified)

    - v and j genes (note that the full field including subgroup + gene is used for matching, i.e. V1 and V1-1 are not considered duplicates)

    - chain

    - region type

    For all other fields (the non-specified sequence type, custom lists, sequence identifier) only the first occurring
    value is kept.

    Note that this means the count value of a sequence with a given sequence identifier might not be the same as before
    removing duplicates, unless count_agg = FIRST is used.

    Specification arguments:

    - filter_sequence_type (:py:obj:`~immuneML.environment.SequenceType.SequenceType`): Whether the sequences should be collapsed on the nucleotide or amino acid level. Valid options are defined by the SequenceType enum.

    - batch_size (int): number of repertoires that can be loaded at the same time (only affects the speed)

    - count_agg (:py:obj:`~immuneML.preprocessing.filters.CountAggregationFunction.CountAggregationFunction`): determines how the sequence counts of duplicate sequences are aggregated. Valid options are defined by the CountAggregationFunction enum.


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter:
                    DuplicateSequenceFilter:
                        # required parameters:
                        filter_sequence_type: AMINO_ACID
                        # optional parameters (if not specified the values bellow will be used):
                        batch_size: 4
                        count_agg: SUM

    """

    @classmethod
    def build_object(cls, **kwargs):
        location = cls.__name__
        ParameterValidator.assert_keys(kwargs.keys(), ["filter_sequence_type", "batch_size", "count_agg"], location,
                                       "DuplicateSequenceFilter")
        ParameterValidator.assert_in_valid_list(kwargs["filter_sequence_type"].upper(), [item.name for item in SequenceType],
                                                location, "filter_sequence_type")
        ParameterValidator.assert_in_valid_list(kwargs["count_agg"].upper(), [item.name for item in CountAggregationFunction], location,
                                                "count_agg")
        ParameterValidator.assert_type_and_value(kwargs["batch_size"], int, location, "batch_size", 1)
        return DuplicateSequenceFilter(filter_sequence_type=SequenceType[kwargs["filter_sequence_type"].upper()],
                                       batch_size=kwargs["batch_size"], count_agg=CountAggregationFunction[kwargs["count_agg"].upper()])

    def __init__(self, filter_sequence_type: SequenceType, batch_size: int, count_agg: CountAggregationFunction, result_path: Path = None):
        super().__init__(result_path)
        self.filter_sequence_type = filter_sequence_type
        self.count_agg = count_agg
        self.batch_size = batch_size

        self.sequence_of_interest = "sequence_aa" if filter_sequence_type == SequenceType.AMINO_ACID else "sequence"
        self.sequence_to_ignore = "sequence" if self.sequence_of_interest == "sequence_aa" else "sequence_aa"

        assert self.sequence_of_interest in Repertoire.FIELDS, f"{DuplicateSequenceFilter.__name__}: {self.sequence_of_interest} not in {Repertoire.FIELDS}"
        assert self.sequence_to_ignore in Repertoire.FIELDS, f"{DuplicateSequenceFilter.__name__}: {self.sequence_of_interest} not in {Repertoire.FIELDS}"

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1) -> RepertoireDataset:
        self.result_path = result_path if result_path is not None else self.result_path

        self.check_dataset_type(dataset, [RepertoireDataset], "DuplicateSequenceFilter")

        processed_dataset = copy.deepcopy(dataset)

        with Pool(self.batch_size) as pool:
            repertoires = pool.map(self._process_repertoire, dataset.repertoires)

        processed_dataset.repertoires = repertoires

        return processed_dataset

    def _prepare_group_by_field(self, columns):
        groupby_fields = copy.deepcopy(list(Repertoire.FIELDS))
        groupby_fields.remove(self.sequence_to_ignore)
        groupby_fields.remove("duplicate_count")
        groupby_fields.remove("sequence_id")
        groupby_fields.remove("cell_id")
        groupby_fields.remove("frame_type")

        for field in set(Repertoire.FIELDS).difference(set(columns)):
            if field in groupby_fields:
                groupby_fields.remove(field)

        return groupby_fields

    def _prepare_agg_dict(self, columns, custom_lists):

        agg_dict = {"sequence_id": "first"}

        if self.sequence_to_ignore in columns:
            agg_dict[self.sequence_to_ignore] = "first"

        if "duplicate_count" in columns:
            agg_dict["duplicate_count"] = self.count_agg.value

        if "cell_id" in columns:
            agg_dict["cell_id"] = "first"

        for key in custom_lists:
            agg_dict[key] = "first"

        return agg_dict

    def _process_repertoire(self, repertoire: Repertoire) -> Repertoire:
        data = repertoire.load_bnp_data().topandas()
        data['duplicate_count'] = [el if el != -1 else pd.NA for el in data['duplicate_count']]
        columns = data.columns

        groupby_fields = self._prepare_group_by_field(columns)
        custom_lists = list(set(columns) - set(Repertoire.FIELDS))
        agg_dict = self._prepare_agg_dict(columns, custom_lists)

        # Chain objects can not be aggregated, convert to strings
        if "chain" in columns:
            data["chain"] = data.chain.tolist()
        else:
            data["chain"] = None

        no_duplicates = data.groupby(groupby_fields, sort=False).agg(agg_dict).reset_index()

        no_duplicates.to_csv(f"{self.result_path}/{repertoire.data_filename.stem}_filtered.tsv", sep='\t', index=False)
        write_yaml(Path(f"{self.result_path}/{repertoire.metadata_filename.stem}_filtered.yaml"), repertoire.metadata)

        return Repertoire(Path(f"{self.result_path}/{repertoire.data_filename.stem}_filtered.tsv"),
                          Path(f"{self.result_path}/{repertoire.metadata_filename.stem}_filtered.yaml"),
                          str(uuid4().hex))

    @staticmethod
    def get_documentation():
        doc = str(DuplicateSequenceFilter.__doc__)

        mapping = {
            "Valid options are defined by the CountAggregationFunction enum.": f"Valid values are: {[e.name.lower() for e in CountAggregationFunction]}.",
            "Valid options are defined by the SequenceType enum.": f"Valid values are: {[e.name.lower() for e in SequenceType]}."
        }

        doc = update_docs_per_mapping(doc, mapping)

        return doc



from pathlib import Path

import pandas as pd

from immuneML.data_model.dataset.ReceptorDataset import ReceptorDataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.PathBuilder import PathBuilder


class GLIPH2Exporter(DataReport):
    """
    Report which exports the receptor data to GLIPH2 format so that it can be directly used in GLIPH2 tool. Currently, the report accepts only
    receptor datasets.

    GLIPH2 publication: Huang H, Wang C, Rubelt F, Scriba TJ, Davis MM. Analyzing the Mycobacterium tuberculosis immune response by T-cell receptor
    clustering with GLIPH2 and genome-wide antigen screening. Nature Biotechnology. Published online April 27,
    2020:1-9. `doi:10.1038/s41587-020-0505-4 <https://www.nature.com/articles/s41587-020-0505-4>`_

    Specification arguments:

    - condition (str): name of the parameter present in the receptor metadata in the dataset; condition can be anything which can be processed in
      GLIPH2, such as tissue type or treatment.

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_gliph2_exporter: # user-defined name
            GLIPH2Exporter:
                condition: epitope # for instance, epitope parameter is present in receptors' metadata with values such as "MtbLys" for Mycobacterium tuberculosis (as shown in the original paper).

    """

    @classmethod
    def build_object(cls, **kwargs):
        return GLIPH2Exporter(**kwargs)

    def __init__(self, dataset: ReceptorDataset = None, result_path: Path = None, name: str = None, condition: str = None, number_of_processes: int = 1):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)
        self.condition = condition

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)
        alpha_chains, beta_chains, trbv, trbj, subject_condition, count = [], [], [], [], [], []
        for index, receptor in enumerate(self.dataset.get_data()):
            alpha_chains.append(receptor.get_chain("alpha").sequence_aa)
            beta_chains.append(receptor.get_chain("beta").sequence_aa)
            trbv.append(receptor.get_chain("beta").metadata.v_call)
            trbj.append(receptor.get_chain("beta").metadata.j_call)
            subject_condition.append(f"{getattr(receptor.metadata, 'subject_id', str(index))}:{receptor.metadata[self.condition]}")
            count.append(receptor.get_chain("beta").metadata.duplicate_count
                         if receptor.get_chain('beta').metadata is not None and receptor.get_chain('beta').metadata.duplicate_count is not None else 1)

        df = pd.DataFrame({"CDR3b": beta_chains, "TRBV": trbv, "TRBJ": trbj, "CDR3a": alpha_chains, "subject:condition": subject_condition,
                           "count": count})
        file_path = self.result_path / "exported_data.tsv"
        df.to_csv(file_path, sep="\t", index=False)

        return ReportResult(self.name,
                            info="Report which exports the receptor data to GLIPH2 format so that it can be directly used in GLIPH2 tool.",
                            output_tables=[ReportOutput(file_path, "exported data in GLIPH2 format")])

    def check_prerequisites(self):
        if isinstance(self.dataset, ReceptorDataset):
            return True
        else:
            return False



import logging

from sklearn.manifold import TSNE as SklearnTSNE

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.ml_methods.dim_reduction.DimRedMethod import DimRedMethod


class TSNE(DimRedMethod):
    """
    t-distributed Stochastic Neighbor Embedding (t-SNE) method which wraps scikit-learn's TSNE. It can be useful for
    visualizing high-dimensional data. Input arguments for the method are the
    same as supported by scikit-learn (see `TSNE scikit-learn documentation
    <https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html#sklearn.manifold.TSNE>`_ for details).

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_tsne: # user-defined name of the dimensionality reduction method
            TSNE: # name of the class
                # arguments as defined by scikit-learn
                n_components: 2
                init: pca

    """
    def __init__(self, name: str = None, **kwargs):
        super().__init__(name)
        self.method_kwargs = kwargs
        self.method = SklearnTSNE(**self.method_kwargs)

    def fit(self, dataset: Dataset):
        self.method.fit(dataset.encoded_data.examples)

    def transform(self, dataset: Dataset):
        logging.warning(f"{TSNE.__name__}: calling transform method of TSNE, but it only supports fit_transform. "
                        f"Fitting the model and returning the transformed data...")
        return self.method.fit_transform(dataset.encoded_data.examples)

    def fit_transform(self, dataset: Dataset):
        return self.method.fit_transform(dataset.encoded_data.examples)


import uuid
from pathlib import Path
from uuid import uuid4

from immuneML.data_model.SequenceSet import SequenceSet
from immuneML.data_model.bnp_util import make_buffer_type_from_dataset_file, write_yaml, read_yaml
from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.encoded_data.EncodedData import EncodedData
from immuneML.data_model.receptor.ElementGenerator import ElementGenerator


class ElementDataset(Dataset):
    """
    This is the base class for ReceptorDataset and SequenceDataset which implements all the functionality for both classes. The only difference between
    these two classes is whether paired or single chain data is stored.
    """

    @classmethod
    def build(cls, dataset_file: Path, types: dict = None, filenames: list = None, **kwargs):
        if not Path(dataset_file).exists():
            metadata = {
                'type_dict': {key: SequenceSet.TYPE_TO_STR[val] for key, val in types.items()},
                'dataset_class': cls.__name__, 'element_class_name': kwargs['element_class_name'],
                'filenames': [str(file) for file in filenames]
            }
            write_yaml(dataset_file, metadata)
        return cls(**{**kwargs, 'dataset_file': dataset_file, 'filenames': filenames})

    def __init__(self, labels: dict = None, encoded_data: EncodedData = None, filenames: list = None,
                 identifier: str = None, dataset_file: Path = None,
                 file_size: int = 100000, name: str = None, element_class_name: str = None,
                 element_ids: list = None, example_weights: list = None,
                 buffer_type=None):
        super().__init__(encoded_data, name, identifier if identifier is not None else uuid4().hex, labels, example_weights)
        self.filenames = filenames if filenames is not None else []
        self.filenames = [Path(filename) for filename in self.filenames]
        if buffer_type is None:
            buffer_type = make_buffer_type_from_dataset_file(Path(dataset_file))
        self.element_generator = ElementGenerator(self.filenames, file_size, element_class_name, buffer_type)
        self.file_size = file_size
        self.element_ids = element_ids
        self.element_class_name = element_class_name
        self.dataset_file = Path(dataset_file)

    def get_data(self, batch_size: int = 10000, return_objects: bool = True):
        self.element_generator.file_list = self.filenames
        return self.element_generator.build_element_generator(return_objects=return_objects)

    def get_batch(self, batch_size: int = 10000):
        self.element_generator.file_list = self.filenames
        return self.element_generator.build_batch_generator()

    def get_filenames(self):
        return self.filenames

    def set_filenames(self, filenames):
        self.filenames = filenames

    def get_example_count(self):
        return len(self.get_example_ids())

    def get_example_ids(self):
        if self.element_ids is None or (isinstance(self.element_ids, list) and len(self.element_ids) == 0):
            self.element_ids = []
            for element in self.get_data():
                self.element_ids.append(str(element.get_id()))
        return self.element_ids

    def get_attribute(self, attribute: str, as_list: bool = False):
        res = self.element_generator.get_attribute(attribute)
        if as_list:
            return res.tolist()
        else:
            return res

    def get_attributes(self, attributes: list, as_list: bool = False) -> dict:
        res = self.element_generator.get_attributes(attributes)
        if as_list:
            return {attr: val.tolist() for attr, val in res.items()}
        else:
            return res

    def make_subset(self, example_indices, path, dataset_type: str):
        """
        Creates a new dataset object with only those examples (receptors or receptor sequences) available which were given by index in example_indices argument.

        Args:
            example_indices (list): a list of indices of examples (receptors or receptor sequences) to use in the new dataset
            path (Path): a path where to store the newly created dataset
            dataset_type (str): a type of the dataset used as a part of the name of the resulting dataset; the values are defined as constants in :py:obj:`~immuneML.data_model.dataset.Dataset.Dataset`

        Returns:

            a new dataset object (ReceptorDataset or SequenceDataset, as the original dataset) which includes only the examples specified under example_indices

        """
        new_dataset_id = uuid.uuid4().hex

        batch_filenames = self.element_generator.make_subset(example_indices, path, dataset_type, new_dataset_id,
                                                             paired=self.element_class_name != 'ReceptorSequence')
        dataset_name = f"{self.name}_split_{dataset_type.lower()}"

        types = read_yaml(self.dataset_file)['type_dict']

        new_dataset = self.__class__.build(labels=self.labels, file_size=self.file_size, filenames=batch_filenames,
                                           element_class_name=self.element_generator.element_class_name,
                                           dataset_file=path / f"{dataset_name}.yaml", types=types,
                                           identifier=new_dataset_id, name=dataset_name)

        # todo check if this is necessary
        original_example_weights = self.get_example_weights()
        if original_example_weights is not None:
            new_dataset.set_example_weights([original_example_weights[i] for i in example_indices])

        return new_dataset

    def get_label_names(self):
        """Returns the list of metadata fields which can be used as labels"""
        return [label for label in list(self.labels.keys()) if
                label not in ['region_type', 'receptor_chains', 'organism']] if self.labels else []

    def clone(self, keep_identifier: bool = False):
        raise NotImplementedError

    def get_data_from_index_range(self, start_index: int, end_index: int):
        return self.element_generator.get_data_from_index_range(start_index, end_index)



from pathlib import Path

from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.receptor.receptor_sequence.Chain import Chain
from immuneML.preprocessing.filters.Filter import Filter


class ChainRepertoireFilter(Filter):
    """
    Removes all repertoires from the RepertoireDataset object which contain at least one sequence
    with chain different than "keep_chain" parameter.
    Note that this filter filters out repertoires, not individual sequences, and can thus only be applied to RepertoireDatasets.

    Since the filter removes repertoires from the dataset (examples in machine learning setting), it cannot be used with :ref:`TrainMLModel`
    instruction. If you want to filter out repertoires including a given chain, see :ref:`DatasetExport` instruction with preprocessing.

    Specification arguments:

    - keep_chain (str): Which chain should be kept, valid values are "TRA", "TRB", "IGH", "IGL", "IGK"


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        preprocessing_sequences:
            my_preprocessing:
                - my_filter:
                    ChainRepertoireFilter:
                        keep_chain: TRB

    """

    def __init__(self, keep_chain, result_path: Path = None):
        super().__init__(result_path)
        self.keep_chain = Chain.get_chain(keep_chain)

    def process_dataset(self, dataset: RepertoireDataset, result_path: Path, number_of_processes=1):
        self.check_dataset_type(dataset, [RepertoireDataset], "ChainRepertoireFilter")
        processed_dataset = dataset.clone()
        self.result_path = result_path if result_path is not None else self.result_path

        repertoires = []
        indices = []
        for index, repertoire in enumerate(dataset.get_data()):
            if all(sequence.metadata.chain == self.keep_chain for sequence in repertoire.sequences):
                repertoires.append(repertoire)
                indices.append(index)

        processed_dataset.repertoires = repertoires
        processed_dataset.metadata_file = self._build_new_metadata(processed_dataset, indices)

        self.check_dataset_not_empty(processed_dataset, "ChainRepertoireFilter")

        return processed_dataset

    def keeps_example_count(self) -> bool:
        return False



from itertools import chain
from pathlib import Path
from typing import List

import math
import numpy as np

from immuneML.data_model.bnp_util import bnp_read_from_file, bnp_write_to_file, make_element_dataset_objects, \
    merge_dataclass_objects
from immuneML.util.ReflectionHandler import ReflectionHandler


class ElementGenerator:

    def __init__(self, file_list: list, file_size: int = 10000, element_class_name: str = "", buffer_type=None):
        self.file_list = file_list
        self.file_lengths = [-1 for _ in range(len(file_list))]
        self.file_size = file_size
        self.element_class_name = element_class_name
        self.buffer_type = buffer_type

    def get_attribute(self, attribute: str):
        elements = []
        for file in self.file_list:
            bnp_data = bnp_read_from_file(file, self.buffer_type)
            elements.append(getattr(bnp_data, attribute))
        return np.concatenate(elements)

    def get_attributes(self, attributes: List[str]):
        elements = {attr: [] for attr in attributes}
        for file in self.file_list:
            bnp_data = bnp_read_from_file(file, self.buffer_type)
            for attribute in attributes:
                elements[attribute].append(getattr(bnp_data, attribute))
        return {attr: np.concatenate(elements[attr]) for attr in attributes}

    def _load_batch(self, current_file: int, return_objects: bool = True):

        element_class = ReflectionHandler.get_class_by_name(self.element_class_name, "data_model")
        assert hasattr(element_class, 'create_from_record'), \
            f"{ElementGenerator.__name__}: cannot load the binary file, the class {element_class.__name__} has no 'create_from_record' method."

        try:
            bnp_data = bnp_read_from_file(self.file_list[current_file], self.buffer_type)
            if return_objects:
                elements = make_element_dataset_objects(bnp_data, self.element_class_name)
            else:
                elements = bnp_data
        except ValueError as error:
            raise ValueError(f'{ElementGenerator.__name__}: an error occurred while creating an object from tsv file. '
                             f'Details: {error}')

        return elements

    def _get_element_count(self, file_index: int):

        if self.file_lengths[file_index] == -1:
            self.file_lengths[file_index] = len(bnp_read_from_file(self.file_list[file_index], self.buffer_type))

        return self.file_lengths[file_index]

    def get_element_count(self):
        for index in range(len(self.file_list)):
            if self.file_lengths[index] == -1:
                self._get_element_count(index)
        return sum(self.file_lengths)

    def build_batch_generator(self, return_objects: bool = True):
        """
        creates a generator which will return one batch of elements at the time

        :param batch_size: how many elements should be returned at once (default 1)
        :return: element generator
        """

        for current_file_index in range(len(self.file_list)):
            batch = self._load_batch(current_file_index, return_objects)
            yield batch

    def build_element_generator(self, return_objects: bool = True):
        """
        creates a generator which will return one element at the time
        :return: element generator
        """
        for current_file_index in range(len(self.file_list)):
            batch = self._load_batch(current_file_index, return_objects)
            if return_objects:
                for element in batch:
                    yield element
            else:
                yield batch

    def make_subset(self, example_indices: list, path: Path, dataset_type: str, dataset_identifier: str, paired: bool = False):
        if example_indices is None or len(example_indices) == 0:
            raise RuntimeError(f"{ElementGenerator.__name__}: no examples were specified to create the dataset subset. "
                               f"Dataset type was {dataset_type}, dataset identifier: {dataset_identifier}.")
        elements = None
        file_count = 1
        tmp_file_size = self.file_size if not paired else self.file_size * 2

        example_indices.sort()

        batch_filenames = self._prepare_batch_filenames(len(example_indices), path, dataset_type, dataset_identifier)

        for index, batch in enumerate(self.build_batch_generator(return_objects=False)):
            extracted_elements = self._extract_elements_from_batch(index, batch, example_indices, paired=paired)
            elements = merge_dataclass_objects([elements, extracted_elements]) if elements else extracted_elements

            if len(elements) >= tmp_file_size or len(elements) == len(example_indices):
                bnp_write_to_file(batch_filenames[file_count - 1], elements[:tmp_file_size])
                file_count += 1
                elements = elements[tmp_file_size:]

        if len(elements) > 0:
            bnp_write_to_file(batch_filenames[file_count - 1], elements)

        return batch_filenames

    def _prepare_batch_filenames(self, example_count: int, path: Path, dataset_type: str, dataset_identifier: str):
        batch_count = math.ceil(example_count / self.file_size)
        digits_count = len(str(batch_count)) + 1
        filenames = [
            path / f"{dataset_identifier}_{dataset_type}_batch{''.join(['0' for _ in range(digits_count - len(str(index)))])}{index}.tsv"
            for index in range(batch_count)]
        return filenames

    def _extract_elements_from_batch(self, index, batch, example_indices, paired: bool = False):
        if paired:
            upper_limit, lower_limit = (index + 1) * len(batch) / 2, index * len(batch) / 2
            batch_indices = list(chain.from_iterable([((ind - lower_limit) * 2, (ind - lower_limit) * 2 + 1)
                                                      for ind in example_indices if lower_limit <= ind < upper_limit]))
        else:
            upper_limit, lower_limit = (index + 1) * len(batch), index * len(batch)
            batch_indices = [ind - lower_limit for ind in example_indices if lower_limit <= ind < upper_limit]

        elements = batch[[int(i) for i in batch_indices]]
        return elements

    def get_data_from_index_range(self, start_index: int, end_index: int, obj_in_two_lines: bool = False):
        elements = []
        start_file_index = start_index // self.file_size
        end_file_index = min(len(self.file_list) - 1, end_index // self.file_size)

        for current_file_index in range(start_file_index, end_file_index + 1):
            batch = self._load_batch(current_file_index)
            i = start_index % self.file_size if current_file_index == start_file_index else 0
            while len(elements) < end_index - start_index + 1 and i < len(batch):
                elements.append(batch[i])
                i += 1

        return elements



import abc
from typing import List

from bionumpy.bnpdataclass import BNPDataClass

from immuneML.environment.SequenceType import SequenceType
from immuneML.simulation.SimConfigItem import SimConfigItem
from immuneML.simulation.implants.Signal import Signal


class SimulationStrategy:

    @abc.abstractmethod
    def process_sequences(self, sequences: BNPDataClass, seqs_per_signal_count: dict, use_p_gens: bool,
                          sequence_type: SequenceType, sim_item: SimConfigItem, all_signals: List[Signal],
                          remove_positives_first: bool, **kwargs) -> BNPDataClass:
        pass


import numpy as np

from immuneML.encodings.EncoderParams import EncoderParams
from immuneML.encodings.word2vec.Word2VecEncoder import Word2VecEncoder
from immuneML.environment.SequenceType import SequenceType
from immuneML.util.KmerHelper import KmerHelper


class W2VSequenceEncoder(Word2VecEncoder):

    def _encode_labels(self, dataset, params: EncoderParams):

        labels = {name: [] for name in params.label_config.get_labels_by_name()}

        for sequence in dataset.get_data():
            for label_name in params.label_config.get_labels_by_name():
                label = sequence.get_attribute(label_name)
                labels[label_name].append(label)

        return np.array([labels[name] for name in labels.keys()])

    def _encode_item(self, item, vectors, sequence_type: SequenceType):
        kmers = KmerHelper.create_kmers_from_sequence(sequence=item, k=self.k, sequence_type=sequence_type)
        sequence_vector = np.zeros(vectors.vector_size)
        for kmer in kmers:
            try:
                word_vector = vectors.get_vector(kmer)
                sequence_vector = np.add(sequence_vector, word_vector)
            except KeyError:
                pass
        return sequence_vector



import logging

import pandas as pd

from immuneML.caching.CacheHandler import CacheHandler
from immuneML.data_model.dataset.ReceptorDataset import ReceptorDataset
from immuneML.data_model.receptor.RegionType import RegionType


class TCRdistHelper:

    @staticmethod
    def compute_tcr_dist(dataset: ReceptorDataset, label_names: list, cores: int = 1):
        return CacheHandler.memo_by_params((('dataset_identifier', dataset.identifier), ("type", "TCRrep")),
                                           lambda: TCRdistHelper._compute_tcr_dist(dataset, label_names, cores))

    @staticmethod
    def _compute_tcr_dist(dataset: ReceptorDataset, label_names: list, cores: int):
        """
        Computes the tcrdist distances by creating a TCRrep object and calling compute_distances() function.

        Parameters `ntrim` and `ctrim` in TCRrep object for CDR3 are adjusted to account for working with IMGT CDR3 definition if IMGT CDR3 was set
        as region_type for the dataset upon importing. `deduplicate` parameter is set to False as we assume that we work with clones in immuneML,
        and not individual receptors.

        Args:
            dataset: receptor dataset for which all pairwise distances between receptors will be computed
            label_names: a list of label names (e.g., specific epitopes) to be used for later classification or reports
            cores: how many cpus to use for computation

        Returns:
            an instance of TCRrep object with computed pairwise distances between all receptors in the dataset

        """
        from tcrdist.repertoire import TCRrep

        df = TCRdistHelper.prepare_tcr_dist_dataframe(dataset, label_names)
        tcr_rep = TCRrep(cell_df=df, chains=['alpha', 'beta'], organism=dataset.labels["organism"], cpus=cores, deduplicate=False,
                         compute_distances=False)

        if 'region_type' not in dataset.labels:
            logging.warning(f"{TCRdistHelper.__name__}: Parameter 'region_type' was not set for dataset {dataset.name}, keeping default tcrdist "
                            f"values for parameters 'ntrim' and 'ctrim'. For more information, see tcrdist3 documentation. To avoid this warning, "
                            f"set the region type when importing the dataset.")
        elif dataset.labels['region_type'] == RegionType.IMGT_CDR3:
            tcr_rep.kargs_a['cdr3_a_aa']['ntrim'] = 2
            tcr_rep.kargs_a['cdr3_a_aa']['ctrim'] = 1
            tcr_rep.kargs_b['cdr3_b_aa']['ntrim'] = 2
            tcr_rep.kargs_b['cdr3_b_aa']['ctrim'] = 1
        elif dataset.labels['region_type'] != RegionType.IMGT_JUNCTION:
            raise RuntimeError(f"{TCRdistHelper.__name__}: TCRdist metric can be computed only if IMGT_CDR3 or IMGT_JUNCTION are used as region "
                               f"types, but for dataset {dataset.name}, it is set to {dataset.labels['region_type']} instead.")

        tcr_rep.compute_distances()

        return tcr_rep

    @staticmethod
    def add_default_allele_to_v_gene(v_gene: str):
        if v_gene is not None and "*" not in v_gene:
            return f"{v_gene}*01"
        else:
            return v_gene

    @staticmethod
    def prepare_tcr_dist_dataframe(dataset: ReceptorDataset, label_names: list) -> pd.DataFrame:
        if len(label_names) > 1:
            raise NotImplementedError(f"TCRdist: multiple labels specified ({str(label_names)[1:-1]}), but only single label binary class "
                                      f"is currently supported in immuneML.")
        label_name = label_names[0]

        subject, epitope, count, v_a_gene, j_a_gene, cdr3_a_aa, v_b_gene, j_b_gene, cdr3_b_aa, clone_id, cdr3_b_nucseq, cdr3_a_nucseq = \
            [], [], [], [], [], [], [], [], [], [], [], []

        for receptor in dataset.get_data():
            subject.append(receptor.metadata["subject"] if "subject" in receptor.metadata else "sub" + receptor.identifier)
            epitope.append(receptor.metadata[label_name])
            count.append(receptor.get_chain("alpha").metadata.duplicate_count
                         if receptor.get_chain("alpha").metadata.duplicate_count == receptor.get_chain("beta").metadata.duplicate_count
                            and receptor.get_chain("beta").metadata.duplicate_count is not None else 1)
            v_a_gene.append(TCRdistHelper.add_default_allele_to_v_gene(receptor.get_chain('alpha').metadata.v_call))
            j_a_gene.append(receptor.get_chain('alpha').metadata.j_call)
            cdr3_a_aa.append(receptor.get_chain('alpha').sequence_aa)
            cdr3_a_nucseq.append(receptor.get_chain("alpha").sequence)
            v_b_gene.append(TCRdistHelper.add_default_allele_to_v_gene(receptor.get_chain('beta').metadata.v_call))
            j_b_gene.append(receptor.get_chain('beta').metadata.v_call)
            cdr3_b_aa.append(receptor.get_chain('beta').sequence_aa)
            cdr3_b_nucseq.append(receptor.get_chain("beta").sequence)
            clone_id.append(receptor.identifier)

        if all(item is not None for item in cdr3_a_nucseq) and all(item is not None for item in cdr3_b_nucseq):
            return pd.DataFrame({"subject": subject, "epitope": epitope, "count": count, "v_a_gene": v_a_gene, "j_a_gene": j_a_gene,
                                 "cdr3_a_aa": cdr3_a_aa, "v_b_gene": v_b_gene, "j_b_gene": j_b_gene, "cdr3_b_aa": cdr3_b_aa, "clone_id": clone_id,
                                 "cdr3_b_nucseq": cdr3_b_nucseq, "cdr3_a_nucseq": cdr3_a_nucseq})
        else:
            return pd.DataFrame({"subject": subject, "epitope": epitope, "count": count, "v_a_gene": v_a_gene, "j_a_gene": j_a_gene,
                                 "cdr3_a_aa": cdr3_a_aa, "v_b_gene": v_b_gene, "j_b_gene": j_b_gene, "cdr3_b_aa": cdr3_b_aa, "clone_id": clone_id})



import json
import warnings

import pandas as pd

from immuneML.IO.dataset_import.DataImport import DataImport
from immuneML.IO.dataset_import.DatasetImportParams import DatasetImportParams
from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.receptor.ChainPair import ChainPair
from immuneML.data_model.receptor.RegionType import RegionType
from immuneML.data_model.receptor.receptor_sequence.SequenceFrameType import SequenceFrameType
from immuneML.data_model.repertoire.Repertoire import Repertoire
from immuneML.util.ImportHelper import ImportHelper
from scripts.specification_util import update_docs_per_mapping


class VDJdbImport(DataImport):
    """
    Imports data in VDJdb format into a Repertoire-, Sequence- or ReceptorDataset.
    RepertoireDatasets should be used when making predictions per repertoire, such as predicting a disease state.
    SequenceDatasets or ReceptorDatasets should be used when predicting values for unpaired (single-chain) and paired
    immune receptors respectively, like antigen specificity.


    Specification arguments:

    - path (str): For RepertoireDatasets, this is the path to a directory with VDJdb files to import. For Sequence- or ReceptorDatasets this path may either be the path to the file to import, or the path to the folder locating one or multiple files with .tsv, .csv or .txt extensions. By default path is set to the current working directory.

    - is_repertoire (bool): If True, this imports a RepertoireDataset. If False, it imports a SequenceDataset or ReceptorDataset. By default, is_repertoire is set to True.

    - metadata_file (str): Required for RepertoireDatasets. This parameter specifies the path to the metadata file. This is a csv file with columns filename, subject_id and arbitrary other columns which can be used as labels in instructions. For setting Sequence- or ReceptorDataset metadata, metadata_file is ignored, see metadata_column_mapping instead.

    - paired (str): Required for Sequence- or ReceptorDatasets. This parameter determines whether to import a SequenceDataset (paired = False) or a ReceptorDataset (paired = True). In a ReceptorDataset, two sequences with chain types specified by receptor_chains are paired together based on the identifier given in the VDJdb column named 'complex.id'.

    - receptor_chains (str): Required for ReceptorDatasets. Determines which pair of chains to import for each Receptor. Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum. If receptor_chains is not provided, the chain pair is automatically detected (only one chain pair type allowed per repertoire).

    - import_illegal_characters (bool): Whether to import sequences that contain illegal characters, i.e., characters that do not appear in the sequence alphabet (amino acids including stop codon '*', or nucleotides). When set to false, filtering is only applied to the sequence type of interest (when running immuneML in amino acid mode, only entries with illegal characters in the amino acid sequence are removed). By default import_illegal_characters is False.

    - import_empty_nt_sequences (bool): imports sequences which have an empty nucleotide sequence field; can be True or False. By default, import_empty_nt_sequences is set to True.

    - import_empty_aa_sequences (bool): imports sequences which have an empty amino acid sequence field; can be True or False; for analysis on amino acid sequences, this parameter should be False (import only non-empty amino acid sequences). By default, import_empty_aa_sequences is set to False.

    - region_type (str): Which part of the sequence to import. By default, this value is set to IMGT_CDR3. This means the first and last amino acids are removed from the CDR3 sequence, as VDJdb uses IMGT junction as CDR3. Specifying any other value will result in importing the sequences as they are. Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.

    - column_mapping (dict): A mapping from VDJdb column names to immuneML's internal data representation. A custom column mapping can be specified here if necessary (for example; adding additional data fields if they are present in the VDJdb file, or using alternative column names). Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS. For VDJdb, this is by default set to:

        .. indent with spaces
        .. code-block:: yaml

                V: v_call
                J: j_call
                CDR3: sequence_aa
                complex.id: sequence_id
                Gene: chain

    - column_mapping_synonyms (dict): This is a column mapping that can be used if a column could have alternative names. The formatting is the same as column_mapping. If some columns specified in column_mapping are not found in the file, the columns specified in column_mapping_synonyms are instead attempted to be loaded. For VDJdb format, there is no default column_mapping_synonyms.

    - metadata_column_mapping (dict): Specifies metadata for Sequence- and ReceptorDatasets. This should specify a mapping similar to column_mapping where keys are VDJdb column names and values are the names that are internally used in immuneML as metadata fields. This means that epitope, epitope_gene and epitope_species can be used as prediction labels for Sequence- and ReceptorDatasets. This parameter can also be used to specify sequence-level metadata columns for RepertoireDatasets, which can be used by reports. To set prediction label metadata for RepertoireDatasets, see metadata_file instead. For VDJdb format, this parameter is by default set to:

        .. indent with spaces
        .. code-block:: yaml

                Epitope: epitope
                Epitope gene: epitope_gene
                Epitope species: epitope_species

    - separator (str): Column separator, for VDJdb this is by default "\\t".


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_vdjdb_dataset:
            format: VDJdb
            params:
                path: path/to/files/
                is_repertoire: True # whether to import a RepertoireDataset
                metadata_file: path/to/metadata.csv # metadata file for RepertoireDataset
                paired: False # whether to import SequenceDataset (False) or ReceptorDataset (True) when is_repertoire = False
                receptor_chains: TRA_TRB # what chain pair to import for a ReceptorDataset
                import_illegal_characters: False # remove sequences with illegal characters for the sequence_type being used
                import_empty_nt_sequences: True # keep sequences even though the nucleotide sequence might be empty
                import_empty_aa_sequences: False # filter out sequences if they don't have sequence_aa set
                # Optional fields with VDJdb-specific defaults, only change when different behavior is required:
                separator: "\\t" # column separator
                region_type: IMGT_CDR3 # what part of the sequence to import
                column_mapping: # column mapping VDJdb: immuneML
                    V: v_call
                    J: j_call
                    CDR3: sequence_aa
                    complex.id: sequence_id
                    Gene: chain
                metadata_column_mapping: # metadata column mapping VDJdb: immuneML
                    Epitope: epitope
                    Epitope gene: epitope_gene
                    Epitope species: epitope_species

    """

    KEY_MAPPING = {
        "subject.id": "subject_id"
    }

    @staticmethod
    def import_dataset(params: dict, dataset_name: str) -> Dataset:
        return ImportHelper.import_dataset(VDJdbImport, params, dataset_name)

    @staticmethod
    def preprocess_dataframe(df: pd.DataFrame, params: DatasetImportParams):
        df["frame_type"] = SequenceFrameType.IN.value
        ImportHelper.junction_to_cdr3(df, params.region_type)
        df.loc[:, "region_type"] = params.region_type.name

        if not params.is_repertoire and params.paired:
            n_single_chains = sum(df["sequence_id"] == "0")
            if n_single_chains > 0:
                df.drop(df.loc[df["sequence_id"] == "0"].index, inplace=True)
                warnings.warn(f"VDJdbImport: {n_single_chains} single chains were removed when trying to create a ReceptorDataset.\n"
                              f"To import all chains as a SequenceDataset, use paired = False")
        else:
            df.loc[df["sequence_id"] == "0", "sequence_id"] = None

        ImportHelper.drop_empty_sequences(df, params.import_empty_aa_sequences, params.import_empty_nt_sequences)
        ImportHelper.drop_illegal_character_sequences(df, params.import_illegal_characters, params.import_with_stop_codon)
        ImportHelper.load_chains(df)

        df["receptor_id"] = df["sequence_id"]
        df["sequence_id"] = VDJdbImport.get_sequence_identifiers(df["sequence_id"], df["chain"])

        df = VDJdbImport.extract_meta_columns(df, params)

        return df

    @staticmethod
    def extract_meta_columns(df: pd.DataFrame, params) -> pd.DataFrame:
        """
        extracts values from meta columns in VDJdb format to separate columns in the data frame, using VDJdbImport.KEY_MAPPING

        Note: the KEY_MAPPING values cannot be manually set, but correspond to the standard names as defined by the AIRR Community

        Args:
            df: data frame of from file[s] in VDJdb which have already been preprocessed
            params: parameters from importing dataset which includes the information on metadata column mapping

        Returns:
            the data frame with additional columns where the metadata (if present) were extracted
        """
        for key, new_key in VDJdbImport.KEY_MAPPING.items():
            df.loc[:, new_key] = ""
        meta_name = params.metadata_column_mapping["Meta"] if params.metadata_column_mapping is not None and "Meta" in params.metadata_column_mapping else "Meta"
        if meta_name in df.columns:
            for index, row in df.iterrows():
                if isinstance(row[meta_name], str):
                    meta = json.loads(row[meta_name])
                    for key, new_key in VDJdbImport.KEY_MAPPING.items():
                        if key in meta:
                            row[new_key] = meta[key]

        return df

    @staticmethod
    def get_sequence_identifiers(receptor_identifiers, chains):
        sequence_identifiers = receptor_identifiers + "_" + chains
        if sequence_identifiers.is_unique:
            return sequence_identifiers
        else:
            counts = sequence_identifiers.value_counts()
            for id, count in counts[counts > 1].items():
                unique_ids = [f"{id}{i}" for i in range(1, count+1)]
                sequence_identifiers.loc[sequence_identifiers == id] = unique_ids
        return sequence_identifiers

    @staticmethod
    def import_receptors(df, params):
        return ImportHelper.import_receptors(df, params)

    @staticmethod
    def get_documentation():
        doc = str(VDJdbImport.__doc__)

        chain_pair_values = str([chain_pair.name for chain_pair in ChainPair])[1:-1].replace("'", "`")
        region_type_values = str([region_type.name for region_type in RegionType])[1:-1].replace("'", "`")
        repertoire_fields = list(Repertoire.FIELDS)
        repertoire_fields.remove("region_type")

        mapping = {
            "Valid values for receptor_chains are the names of the :py:obj:`~immuneML.data_model.receptor.ChainPair.ChainPair` enum.": f"Valid values are {chain_pair_values}.",
            "Valid values for region_type are the names of the :py:obj:`~immuneML.data_model.receptor.RegionType.RegionType` enum.": f"Valid values are {region_type_values}.",
            "Valid immuneML fields that can be specified here are defined by Repertoire.FIELDS": f"Valid immuneML fields that can be specified here are {repertoire_fields}."
        }
        doc = update_docs_per_mapping(doc, mapping)
        return doc






# quality: gold
import logging
import warnings
from typing import List

from immuneML.environment.Label import Label
from immuneML.util.ParameterValidator import ParameterValidator


class LabelConfiguration:
    """
    Class that encapsulates labels and transformers for the labels.
    Supports two types of labels: CLASSIFICATION and REGRESSION (as defined in LabelType class)
    """

    def __init__(self, labels: list = None):

        assert labels is None or all(isinstance(label, Label) for label in labels), \
            "LabelConfiguration: all labels should be instances of Label class."

        self._labels = {label.name: label for label in labels} if labels is not None else {}

    def add_label(self, label_name: str, values: list = None, auxiliary_labels: list = None, positive_class=None):

        vals = list(values) if values else None

        if label_name in self._labels and self._labels[label_name] is not None and len(self._labels[label_name]) > 0:
            warnings.warn("Label " + label_name + " has already been set. Overriding existing values...", Warning)

        if positive_class is not None:
            if all(isinstance(val, str) for val in vals) and not isinstance(positive_class, str):
                positive_class = str(positive_class)
            ParameterValidator.assert_in_valid_list(positive_class, vals, Label.__name__, 'positive_class')
        else:
            positive_class = self._get_default_positive_class(vals)
            logging.info(f"LabelConfiguration: No positive label class was set. "
                         f"Setting default positive class '{positive_class}' for label {label_name}")

        self._labels[label_name] = Label(label_name, vals, auxiliary_labels, positive_class)

    def _get_default_positive_class(self, classes):
        """Returns the default positive class when a class pair is given where the positive class is obvious (0,
        1; true, false)"""

        if len(classes) != 2:
            return None
        if classes[0] == True and classes[1] == False:
            return classes[0]
        if classes[1] == True and classes[0] == False:
            return classes[1]

        for positive_str, negative_str in [("1", "0"), ("true", "false"), ("positive", "negative"), ("+", "-")]:
            if set(classes) == {positive_str, negative_str}:
                return positive_str
            if set(classes) == {positive_str.upper(), negative_str.upper()}:
                return positive_str.upper()
            if set(classes) == {positive_str.title(), negative_str.title()}:
                return positive_str.title()

        return sorted(classes)[0]

    def get_labels_by_name(self):
        return sorted(list(self._labels.keys()))

    def get_label_values(self, label_name: str):
        assert label_name in self._labels, label_name + " is not in the list of labels, so there is no information on the values."
        return self._labels[label_name].values

    def get_label_count(self):
        return len(self._labels.keys())

    def get_auxiliary_labels(self, label_name: str):
        return self._labels[label_name].auxiliary_label_names

    def get_label_object(self, label_name: str) -> Label:
        return self._labels[label_name]

    def get_label_objects(self) -> List[Label]:
        return list(self._labels.values())



from pathlib import Path

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.data_model.dataset.ReceptorDataset import ReceptorDataset
from immuneML.data_model.dataset.RepertoireDataset import RepertoireDataset
from immuneML.data_model.dataset.SequenceDataset import SequenceDataset
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.data_reports.DataReport import DataReport
from immuneML.util.PathBuilder import PathBuilder


class SimpleDatasetOverview(DataReport):
    """
    Generates a simple text-based overview of the properties of any dataset, including the dataset name, size, and metadata labels.

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        reports:
            my_overview: SimpleDatasetOverview

    """
    UNKNOWN_CHAIN = "unknown"

    def __init__(self, dataset: Dataset = None, result_path: Path = None, number_of_processes: int = 1, name: str = None):
        super().__init__(dataset=dataset, result_path=result_path, number_of_processes=number_of_processes, name=name)

    @classmethod
    def build_object(cls, **kwargs):
        return SimpleDatasetOverview(**kwargs)

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)

        text_path = self.result_path / "dataset_description.txt"

        dataset_name = self.dataset.name if self.dataset.name is not None else self.dataset.identifier

        output_text = self._get_generic_dataset_text()

        if isinstance(self.dataset, RepertoireDataset):
            output_text += self._get_repertoire_dataset_text()
        elif isinstance(self.dataset, ReceptorDataset):
            output_text += self._get_receptor_dataset_text()
        elif isinstance(self.dataset, SequenceDataset):
            output_text += self._get_sequence_dataset_text()

        text_path.write_text(output_text)

        return ReportResult(name=self.name,
                            info=f"A simple overview of the properties of dataset {self.dataset.name}",
                            output_text=[ReportOutput(text_path, f"Description of dataset {dataset_name}")])

    def _get_generic_dataset_text(self):
        element_name = type(self.dataset).__name__.replace("Dataset", "s").lower()

        output_text = f"Dataset name: {self.dataset.name}\n" \
                      f"Dataset identifier: {self.dataset.identifier}\n" \
                      f"Dataset type: {type(self.dataset).__name__}\n" \
                      f"Dataset size: {self.dataset.get_example_count()} {element_name}\n" \
                      f"Labels available for classification:"

        if len(self.dataset.get_label_names()) == 0:
            output_text += " None"
        else:
            for label in self.dataset.get_label_names():
                output_text += "\n - " + label

        return output_text

    def _get_repertoire_dataset_text(self):
        output_text = f"\nmetadata file location: {self.dataset.metadata_file}\n"

        output_text += "\n\nProperties per repertoire:\n"
        for repertoire in self.dataset.repertoires:
            output_text += f"- Name: {repertoire.data_filename.name}\n"
            output_text += f"  Number of sequences: {repertoire.get_element_count()}\n"

            chains = [chain.value if chain else SimpleDatasetOverview.UNKNOWN_CHAIN for chain in set(repertoire.get_chains())]
            if len(chains) == 1:
                output_text += f"  Chain type: {chains[0]}\n"
            else:
                output_text += f"  Chain types: {','.join(chains)}\n"

        return output_text

    def _get_receptor_dataset_text(self):
        receptor_types = list(set([type(receptor).__name__ for receptor in self.dataset.get_data()]))

        if len(receptor_types) > 1:
            output_text = "\nReceptor types: " + ",".join(receptor_types)
        else:
            output_text = "\nReceptor type: " + receptor_types[0]

        return output_text

    def _get_sequence_dataset_text(self):
        chains = list(set([sequence.get_attribute("chain") for sequence in self.dataset.get_data()]))
        chains = [chain.value if chain else SimpleDatasetOverview.UNKNOWN_CHAIN for chain in chains]

        if len(chains) > 1:
            output_text = "\nChain types: " + ",".join(chains)
        else:
            output_text = "\nChain type: " + chains[0]

        return output_text


import logging
from collections import Counter
from pathlib import Path
from typing import Tuple

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib_venn import venn2

from immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder import CompAIRRSequenceAbundanceEncoder
from immuneML.encodings.abundance_encoding.KmerAbundanceEncoder import KmerAbundanceEncoder
from immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder import SequenceAbundanceEncoder
from immuneML.environment.Label import Label
from immuneML.hyperparameter_optimization.states.TrainMLModelState import TrainMLModelState
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.train_ml_model_reports.TrainMLModelReport import TrainMLModelReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class ReferenceSequenceOverlap(TrainMLModelReport):
    """
    The ReferenceSequenceOverlap report compares a list of disease-associated sequences (or k-mers) produced by the
    :py:obj:`~immuneML.encodings.abundance_encoding.SequenceAbundanceEncoder.SequenceAbundanceEncoder`,
    :py:obj:`~immuneML.encodings.abundance_encoding.CompAIRRSequenceAbundanceEncoder.CompAIRRSequenceAbundanceEncoder` or
    :py:obj:`~immuneML.encodings.abundance_encoding.KmerAbundanceEncoder.KmerAbundanceEncoder` to
    a list of reference sequences. It outputs a Venn diagram and a list of sequences found both in the encoder and reference list.

    The report compares the sequences by their sequence content and the additional comparison_attributes (such as V or J gene), as specified by the user.

    Specification arguments:

    - reference_path (str): path to the reference file in csv format which contains one entry per row and has columns that correspond to the attributes
      listed under comparison_attributes argument

    - comparison_attributes (list): list of attributes to use for comparison; all of them have to be present in the reference file where they should
      be the names of the columns

    - label (str): name of the label for which the reference sequences/k-mers should be compared to the model; if none, it takes the one label from the
      instruction; if it is none and multiple labels were specified for the instruction, the report will not be generated


    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        reports: # the report is defined with all other reports under definitions/reports
            my_reference_overlap_report:
                ReferenceSequenceOverlap:
                    reference_path: reference_sequences.csv  # example usage with SequenceAbundanceEncoder or CompAIRRSequenceAbundanceEncoder
                    comparison_attributes:
                        - sequence_aa
                        - v_call
                        - j_call
            my_reference_overlap_report_with_kmers:
                ReferenceSequenceOverlap:
                    reference_path: reference_kmers.csv  # example usage with KmerAbundanceEncoder
                    comparison_attributes:
                        - k-mer

    """

    @classmethod
    def build_object(cls, **kwargs):

        ParameterValidator.assert_keys(kwargs.keys(), ['reference_path', 'comparison_attributes', 'name', 'label'], ReferenceSequenceOverlap.__name__,
                                       f"reports: {kwargs['name'] if 'name' in kwargs else ''}")

        kwargs['reference_path'] = Path(kwargs['reference_path'])

        assert kwargs['reference_path'].is_file(), f"{ReferenceSequenceOverlap.__name__}: 'reference_path' for report {kwargs['name']} is not " \
                                                         f"a valid file path."

        reference_sequences_df = pd.read_csv(kwargs['reference_path'])
        attributes = reference_sequences_df.columns.tolist()

        ParameterValidator.assert_keys_present(expected_values=kwargs['comparison_attributes'], values=attributes,
                                               location=ReferenceSequenceOverlap.__name__,
                                               parameter_name='columns in file under reference_path')

        return ReferenceSequenceOverlap(**kwargs)

    def __init__(self, reference_path: Path = None, comparison_attributes: list = None, name: str = None, state: TrainMLModelState = None,
                 result_path: Path = None, label: Label = None, number_of_processes: int = 1):
        super().__init__(name=name, state=state, label=label, result_path=result_path, number_of_processes=number_of_processes)
        self.reference_path = reference_path
        self.comparison_attributes = comparison_attributes

    def _generate(self) -> ReportResult:

        figures, tables = [], []

        PathBuilder.build(self.result_path)

        if ReferenceSequenceOverlap._check_encoder_class(self.state.optimal_hp_items[self.label.name].encoder):
            figure, data = self._compute_optimal_model_overlap()
            figures.append(figure)
            tables.append(data)

        for assessment_state in self.state.assessment_states:
            encoder = assessment_state.label_states[self.label.name].optimal_assessment_item.encoder
            if ReferenceSequenceOverlap._check_encoder_class(encoder):
                figure_filename = self.result_path / f"assessment_split_{assessment_state.split_index + 1}_model_vs_reference_overlap_{self.label.name}.pdf"
                df_filename = self.result_path / f"assessment_split_{assessment_state.split_index + 1}_overlap_sequences_{self.label.name}"
                figure, data = self._compute_model_overlap(figure_filename, df_filename, encoder,
                                                           f"overlap sequences between the model for assessment split "
                                                           f"{assessment_state.split_index + 1} and reference list")
                figures.append(figure)
                tables.append(data)

        return ReportResult(self.name,
                            info="A Venn diagram between the list of disease-associated sequences produced by the SequenceAbundance encoder and a list of reference receptor sequences, and a file containing the overlapping sequences.",
                            output_figures=figures, output_tables=tables)

    @staticmethod
    def _check_encoder_class(encoder):
        return any(isinstance(encoder, cls) for cls in [SequenceAbundanceEncoder, CompAIRRSequenceAbundanceEncoder, KmerAbundanceEncoder])

    def check_prerequisites(self):

        valid = True

        if self.label is None:
            if len(self.state.label_configuration.get_labels_by_name()) != 1:

                logging.warning(f"{ReferenceSequenceOverlap.__name__}: label parameter for report {self.name} is None and it could not be inferred "
                                f"from other information available in the report. It can be inferred automatically if there is only one label "
                                f"specified in the analysis, but got {self.state.label_configuration.get_labels_by_name()} instead. Skipping this "
                                f"report...")
                valid = False
            else:
                self.label = self.state.label_configuration.get_label_objects()[0]

        return valid

    def _compute_optimal_model_overlap(self) -> Tuple[ReportOutput, ReportOutput]:

        filename = self.result_path / f"optimal_model_vs_reference_overlap_{self.label.name}.pdf"
        df_filename = self.result_path / f"overlap_sequences_{self.label.name}.csv"
        encoder = self.state.optimal_hp_items[self.label.name].encoder

        return self._compute_model_overlap(filename, df_filename, encoder,
                                           f"overlap sequences between the reference and the optimal model for label {self.label.name}")

    def _compute_model_overlap(self, figure_filename, df_filename, encoder, name):

        reference_sequences_df = pd.read_csv(self.reference_path, usecols=self.comparison_attributes)
        reference_sequences = list(reference_sequences_df.to_records(index=False))
        attributes = reference_sequences_df.columns.tolist()

        model_sequences = self._extract_from_model(encoder)

        overlap_sequences = [sequence for sequence in model_sequences if sequence in reference_sequences]
        count_overlap = len(overlap_sequences)
        count_ref_only = len([sequence for sequence in reference_sequences if sequence not in model_sequences])
        count_model_only = len([sequence for sequence in model_sequences if sequence not in reference_sequences])

        self._make_venn_diagram(count_ref_only, count_overlap, count_model_only, 'reference', 'model', figure_filename)
        figure = ReportOutput(figure_filename, name)

        pd.DataFrame.from_records(overlap_sequences, columns=attributes).to_csv(df_filename, index=False)
        data = ReportOutput(df_filename, name)

        return figure, data

    def _extract_from_model(self, encoder):

        model_sequences_df = pd.read_csv(getattr(encoder, "relevant_sequence_path"))
        model_attributes = model_sequences_df.columns.tolist()
        assert all(attribute in self.comparison_attributes for attribute in model_attributes), \
            f"{ReferenceSequenceOverlap.__name__}: comparison attributes from the report {self.name} ({self.comparison_attributes}) and from the optimal " \
            f"encoding {encoder.name} ({model_attributes}) do not match."

        return list(model_sequences_df[self.comparison_attributes].to_records(index=False))

    def _make_venn_diagram(self, count_ref_only: int, count_overlap: int, count_model_only: int, label_reference: str, label_model: str,
                           filename: str):
        subsets = Counter({"01": count_model_only, "10": count_ref_only, "11": count_overlap})
        diagram = venn2(subsets=subsets, set_labels=(label_reference, label_model), set_colors=('#72AAA1', '#E5B9AD'), alpha=0.8)
        for index in subsets:
            if subsets[index] == 0 and diagram.get_label_by_id(index) is not None:
                diagram.get_label_by_id(index).set_text("")
        plt.savefig(filename)
        plt.clf()



from pathlib import Path

import logging
import plotly.express as px
import pandas as pd
from multiprocessing import Pool
from functools import partial


from immuneML.data_model.dataset import SequenceDataset
from immuneML.encodings.motif_encoding.PositionalMotifHelper import PositionalMotifHelper
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.encodings.motif_encoding.MotifEncoder import MotifEncoder
from immuneML.reports.encoding_reports.EncodingReport import EncodingReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class NonMotifSequenceSimilarity(EncodingReport):
    """
    Plots the similarity of positions outside the motifs of interest. This report can be used to investigate if the
    motifs of interest as determined by the :py:obj:`~immuneML.encodings.motif_encoding.MotifEncoder.MotifEncoder`
    have a tendency occur in sequences that are naturally very similar or dissimilar.

    For each motif, the subset of sequences containing the motif is selected, and the hamming distances are computed
    between all sequences in this subset. Finally, a plot is created showing the distribution of hamming distances
    between the sequences containing the motif. For motifs occurring in sets of very similar sequences, this distribution
    will lean towards small hamming distances. Likewise, for motifs occurring in a very diverse set of sequences, the
    distribution will lean towards containing more large hamming distances.


    Specification arguments:

        motif_color_map (dict): An optional mapping between motif sizes and colors. If no mapping is given, default colors will be chosen.


    YAML specification example:

    .. indent with spaces
    .. code-block:: yaml

        my_motif_sim:
            NonMotifSimilarity:
                motif_color_map:
                    3: "#66C5CC"
                    4: "#F6CF71"
                    5: "#F89C74"

    """

    @classmethod
    def build_object(cls, **kwargs):
        if "motif_color_map" in kwargs:
            ParameterValidator.assert_type_and_value(kwargs["motif_color_map"], dict, NonMotifSequenceSimilarity.__name__, "motif_color_map")
            kwargs["motif_color_map"] = {str(key): value for key, value in kwargs["motif_color_map"].items()}

        return NonMotifSequenceSimilarity(**kwargs)

    def __init__(self, dataset: SequenceDataset = None, result_path: Path = None,
                 motif_color_map: dict = None, name: str = None, number_of_processes: int = 1):
        super().__init__(dataset=dataset, result_path=result_path, name=name, number_of_processes=number_of_processes)
        self.sequence_length = 0
        self.motif_color_map = motif_color_map

    def _generate(self):
        PathBuilder.build(self.result_path)

        raw_counts = self.get_hamming_distance_counts()
        plotting_data = self.get_plotting_data(raw_counts)

        table_1 = self._write_output_table(raw_counts, self.result_path / "sequence_hamming_distances_raw.tsv",
                                                "Hamming distances between sequences sharing the same motif, raw counts.")
        table_2 = self._write_output_table(plotting_data, self.result_path / "sequence_hamming_distances_percentage.tsv",
                                                "Hamming distances between sequences sharing the same motif, expressed as percentage.")

        output_figure = self._safe_plot(plotting_data=plotting_data)

        return ReportResult(
            name=self.name,
            output_figures=[output_figure],
            output_tables=[table_1, table_2],
        )

    def get_hamming_distance_counts(self):
        np_sequences = PositionalMotifHelper.get_numpy_sequence_representation(self.dataset)
        self.sequence_length = len(np_sequences[0])

        raw_counts = pd.DataFrame([self._make_hamming_distance_hist_for_motif(motif_presence, np_sequences)
                                   for motif_presence in self.dataset.encoded_data.examples.T])

        ### Original code with multiprocessing (fails with bionumpy + pickle error?)
        # with Pool(processes=self.number_of_processes) as pool:
        #     partial_func = partial(self._make_hamming_distance_hist_for_motif,  np_sequences=np_sequences)
        #     raw_counts = pd.DataFrame(pool.map(partial_func, self.dataset.encoded_data.examples.T))

        raw_counts["motif"] = self.dataset.encoded_data.feature_names

        return raw_counts
    def _make_hamming_distance_hist_for_motif(self, motif_presence, np_sequences):
        positive_sequences = np_sequences[motif_presence]
        return self._make_hamming_distance_hist(positive_sequences)
    def _make_hamming_distance_hist(self, sequences):
        counts = {i: 0 for i in range(self.sequence_length)}
        for dist in self._calculate_all_hamming_distances(sequences):
            counts[dist] += 1

        return counts
    def _calculate_all_hamming_distances(self, sequences):
        for i in range(len(sequences)):
            for j in range(i + 1, len(sequences)):
                yield sum(sequences[i] != sequences[j])

    def get_plotting_data(self, raw_counts):
        motif_col = raw_counts.loc[:,"motif"]
        counts = raw_counts.loc[:, raw_counts.columns != "motif"]

        total = counts.apply(sum, axis=1)
        plotting_data = counts.div(total, axis=0)
        plotting_data["motif"] = motif_col
        plotting_data = plotting_data[total > 0]
        plotting_data = plotting_data.loc[::-1]
        plotting_data = pd.melt(plotting_data, id_vars=["motif"], var_name="Hamming", value_name="Percentage")

        plotting_data["Hamming"] = plotting_data["Hamming"].astype(str)
        plotting_data["motif_size"] = plotting_data["motif"].apply(lambda motif: len(motif.split("-")[0].split("&")))

        return plotting_data

    def _plot(self, plotting_data) -> ReportOutput:
        if self.motif_color_map is not None:
            color_discrete_map = self.motif_color_map
            color_discrete_sequence = None
        else:
            color_discrete_map = None
            color_discrete_sequence = px.colors.sequential.Sunsetdark

        plotting_data["motif_size"] = plotting_data["motif_size"].astype(str)

        fig = px.line(plotting_data, x='Hamming', y='Percentage', markers=True, line_group='motif', color="motif_size",
                      template='plotly_white', color_discrete_sequence=color_discrete_sequence, color_discrete_map=color_discrete_map,
                      labels={"Percentage": "Percentage of sequence pairs containing motif",
                              "Hamming": "Hamming distance between sequences",
                              "motif_size": "Motif length<br>(number of<br>amino acids)"})

        fig.layout.yaxis.tickformat = ',.0%'

        fig.update_traces(opacity=0.7)

        fig.update_layout(
            font=dict(
                size=14,
            )
        )

        output_path = self.result_path / "sequence_hamming_distances.html"

        fig.write_html(str(output_path))

        return ReportOutput(path=output_path, name="Hamming distances between sequences sharing the same motif")

    def check_prerequisites(self):
        valid_encodings = [MotifEncoder.__name__]

        if self.dataset.encoded_data is None or self.dataset.encoded_data.info is None:
            logging.warning(f"{self.__class__.__name__}: the dataset is not encoded, skipping this report...")
            return False
        elif self.dataset.encoded_data.encoding not in valid_encodings:
            logging.warning(f"{self.__class__.__name__}: the dataset encoding ({self.dataset.encoded_data.encoding}) was not in the list of valid "
                            f"encodings ({valid_encodings}), skipping this report...")
            return False
        else:
            return True



import shutil
from pathlib import Path

import pandas as pd
import torch
from plotly.subplots import make_subplots
import plotly.graph_objects as go
import plotly.express as px
from sklearn.decomposition import PCA

from immuneML.data_model.dataset.Dataset import Dataset
from immuneML.ml_methods.generative_models.GenerativeModel import GenerativeModel
from immuneML.ml_methods.generative_models.SimpleVAE import SimpleVAE
from immuneML.reports.ReportOutput import ReportOutput
from immuneML.reports.ReportResult import ReportResult
from immuneML.reports.gen_model_reports.GenModelReport import GenModelReport
from immuneML.util.ParameterValidator import ParameterValidator
from immuneML.util.PathBuilder import PathBuilder


class VAESummary(GenModelReport):
    """
    This report provides the summary of the train SimpleVAE and shows the following:

    - plots of the latent space after applying PCA to reduce the data to 2 dimensions, highlighted by V and J gene
    - plots the histogram for each latent dimension
    - plots loss per epoch

    Arguments:

        dim_dist_cols (int): how many columns to use to plot the histograms of latent dimensions (either this or dim_dist_rows has to be set, or both)

        dim_dist_rows (int): how many rows to use to plot the histogram of latent dimensions (either this or dim_dist_cols has to be set, or both)

    YAML specification:

    .. indent with spaces
    .. code-block:: yaml

        my_vae_summary:
          VAESummary:
            dim_dist_cols: 4
            dim_dist_rows: None

    """
    @classmethod
    def build_object(cls, **kwargs):
        name = kwargs["name"] if "name" in kwargs else "VAESummary"

        ParameterValidator.assert_keys_present(list(kwargs.keys()), ['dim_dist_cols', 'dim_dist_rows'],
                                               VAESummary.__name__, 'parameters')
        assert ((kwargs['dim_dist_cols'] is not None and isinstance(kwargs['dim_dist_cols'], int)) or
                (kwargs['dim_dist_rows'] is not None and isinstance(kwargs['dim_dist_rows'], int))), \
            f"{cls.__name__}: at least one of dim_dist_cols or dim_dist_rows has to be set."

        return VAESummary(name=name, dim_dist_cols=kwargs['dim_dist_cols'], dim_dist_rows=kwargs['dim_dist_rows'])

    def __init__(self, dim_dist_rows: int, dim_dist_cols: int, dataset: Dataset = None, model: GenerativeModel = None,
                 result_path: Path = None, name: str = None):
        super().__init__(dataset, model, result_path, name)
        self.dim_dist_rows = dim_dist_rows
        self.dim_dist_cols = dim_dist_cols

    def _generate(self) -> ReportResult:
        PathBuilder.build(self.result_path)

        latent_space_table_out = self._prepare_latent_space()
        train_progress_table_out = self._prepare_training_progress()

        latent_space_fig_out = self._safe_plot(output_written=True, plot_callable='_plot_latent_space',
                                               latent_space_table=latent_space_table_out)
        latent_dim_dist_fig_out = self._safe_plot(output_written=True,
                                                  plot_callable='_plot_latent_dimension_distributions')
        train_progress_fig_out = self._safe_plot(output_written=True, plot_callable='_plot_training_progress',
                                                 training_progress_table=train_progress_table_out)

        output_figures = [latent_space_fig_out, latent_dim_dist_fig_out, train_progress_fig_out]
        output_figures = [el for el in output_figures if el is not None]

        result = ReportResult(name=self.name, info='Summary of the fitted VAE model.', output_figures=output_figures,
                              output_tables=[latent_space_table_out, train_progress_table_out])
        return result

    def _prepare_latent_space(self) -> ReportOutput:
        data_loader = self.model.encode_dataset(self.dataset, self.dataset.get_example_count(), False)

        for data in data_loader:
            with torch.no_grad():
                cdr3_input, v_gene_input, j_gene_input = data
                embeddings = self.model.model.encoding_func(cdr3_input, v_gene_input, j_gene_input)

        pca = PCA(n_components=2)
        embeddings = pca.fit_transform(embeddings.numpy())

        df = pd.DataFrame(data=embeddings, columns=['PC1', 'PC2'])
        df['v_gene'] = [v_call.split("*")[0] for v_call in self.dataset.get_attribute('v_call', True)]
        df['j_gene'] = [j_call.split("*")[0] for j_call in self.dataset.get_attribute('j_call', True)]

        path = self.result_path / 'latent_space_2_component_PCA.csv'

        df.to_csv(str(path), index=False)

        return ReportOutput(path, f'principal component analysis on the data embedded into '
                                  f'{self.model.latent_dim} dimensional space')

    def _prepare_training_progress(self) -> ReportOutput:
        path = self.result_path / 'training_losses.csv'
        shutil.copyfile(str(self.model.loss_path), str(path))
        return ReportOutput(path, 'Loss per epoch')

    def _plot_latent_space(self, latent_space_table: ReportOutput) -> ReportOutput:

        encoded_data = pd.read_csv(latent_space_table.path)

        fig = make_subplots(rows=2, cols=1, shared_xaxes=True, shared_yaxes=True, subplot_titles=['V genes', 'J genes'],
                            x_title='PC1', y_title='PC2')
        for gene in self.model.unique_v_genes:
            tmp_df = encoded_data[encoded_data['v_gene'] == gene]
            fig.add_trace(
                go.Scatter(x=tmp_df['PC1'].values, y=tmp_df['PC2'].values, mode='markers', opacity=0.5, name=gene,
                           marker=dict(colorscale='Viridis', line_width=1), legendgroup=1,
                           legendgrouptitle_text='V genes'), row=1, col=1)
        for gene in self.model.unique_j_genes:
            tmp_df = encoded_data[encoded_data['j_gene'] == gene]
            fig.add_trace(
                go.Scatter(x=tmp_df['PC1'].values, y=tmp_df['PC2'].values, mode='markers', opacity=0.5, name=gene,
                           marker=dict(colorscale='Viridis', line_width=1),
                           legendgroup=2, legendgrouptitle_text='J genes'), row=2, col=1)

        fig.update_layout(template='plotly_white', legend_tracegroupgap=30)
        fig.write_html(self.result_path / 'latent_space_PCA.html')

        return ReportOutput(self.result_path / 'latent_space_PCA.html',
                            f'principal component analysis on the data embedded into {self.model.latent_dim} '
                            f'dimensional space')

    def _plot_latent_dimension_distributions(self) -> ReportOutput:
        data_loader = self.model.encode_dataset(self.dataset, self.dataset.get_example_count(), False)

        for data in data_loader:
            with torch.no_grad():
                cdr3_input, v_gene_input, j_gene_input = data
                embeddings = self.model.model.encoding_func(cdr3_input, v_gene_input, j_gene_input)

        if self.dim_dist_rows is None:
            self.dim_dist_rows = int(self.model.latent_dim / self.dim_dist_cols)
        elif self.dim_dist_cols is None:
            self.dim_dist_cols = int(self.model.latent_dim / self.dim_dist_rows)

        assert self.dim_dist_rows * self.dim_dist_cols == self.model.latent_dim, \
            (f"{VAESummary.__name__}: cannot plot latent dimension distribution since dim_dist_cols and dim_dist_rows "
             f"parameters do not match the latent dimension of the VAE.")

        fig = make_subplots(rows=self.dim_dist_rows, cols=self.dim_dist_cols, shared_xaxes=True, shared_yaxes=True)

        i = 0
        for row in range(self.dim_dist_rows):
            for col in range(self.dim_dist_cols):
                fig.add_trace(go.Histogram(x=embeddings[:, i].numpy(), name=f'dim {i + 1}'),
                              row=row + 1, col=col + 1)
                i += 1

        fig.update_layout(template='plotly_white')
        fig.write_html(self.result_path / 'latent_dim_dist.html')

        return ReportOutput(path=self.result_path / 'latent_dim_dist.html',
                            name=f'latent dimension distribution for examples from dataset {self.dataset.name}')

    def _plot_training_progress(self, training_progress_table: ReportOutput) -> ReportOutput:
        df = pd.read_csv(str(training_progress_table.path))
        fig = px.line(df, x='epoch', y='loss', markers=True)
        fig.update_layout(template='plotly_white')
        fig.write_html(self.result_path / 'loss_per_epoch.html')

        return ReportOutput(path=self.result_path / 'loss_per_epoch.html',
                            name='loss per epoch')

    def check_prerequisites(self) -> bool:
        return isinstance(self.model, SimpleVAE)



import abc
import pickle
import shutil
from pathlib import Path
from typing import List

from immuneML.IO.dataset_export.ImmuneMLExporter import ImmuneMLExporter
from immuneML.encodings.EncoderParams import EncoderParams


class DatasetEncoder(metaclass=abc.ABCMeta):
    """

    YAML specification:

        encodings:
            e1: <encoder_class> # encoding without parameters

            e2:
                <encoder_class>: # encoding with parameters
                    parameter: value
    """

    @staticmethod
    @abc.abstractmethod
    def build_object(dataset, **params):
        pass

    @abc.abstractmethod
    def encode(self, dataset, params: EncoderParams):
        pass

    @staticmethod
    def load_encoder(encoder_file: Path):
        with encoder_file.open("rb") as file:
            encoder = pickle.load(file)
        return encoder

    @staticmethod
    def load_attribute(encoder, encoder_file: Path, attribute: str):
        if encoder_file is not None:
            file_path = encoder_file.parent / getattr(encoder, attribute).name
            setattr(encoder, attribute, file_path)
            assert getattr(encoder, attribute).is_file(), f"{type(encoder).__name__}: could not load {attribute} from {getattr(encoder, attribute)}."
        return encoder

    @staticmethod
    def store_encoder(encoder, encoder_file: Path):
        with encoder_file.open("wb") as file:
            pickle.dump(encoder, file)

        encoder_dir = encoder_file.parent
        for file in encoder.get_additional_files():
            shutil.copy(file, encoder_dir / file.name)

        return encoder_file

    @staticmethod
    def get_additional_files() -> List[str]:
        """Should return a list with all the files that need to be stored when storing the encoder."""
        return []

    def set_context(self, context: dict):
        return self

    def store(self, encoded_dataset, params: EncoderParams):
        ImmuneMLExporter.export(encoded_dataset, params.result_path)



from immuneML.data_model.receptor.Receptor import Receptor
from immuneML.data_model.receptor.receptor_sequence.ReceptorSequence import ReceptorSequence


class BCKReceptor(Receptor):
    FIELDS = {'heavy': object, 'kappa': object, 'identifier': str, 'metadata': dict}

    @classmethod
    def get_record_names(cls):
        return ['heavy_' + name for name in ReceptorSequence.get_record_names()] \
               + ['kappa_' + name for name in ReceptorSequence.get_record_names()] \
               + [name for name in cls.FIELDS if name not in ['heavy', 'kappa']]

    def __init__(self, heavy: ReceptorSequence = None, kappa: ReceptorSequence = None, metadata: dict = None, identifier: str = None):
        super().__init__(metadata=metadata, identifier=identifier)
        self.heavy = heavy
        self.kappa = kappa

    def get_chains(self):
        return ["heavy", "kappa"]



from pathlib import Path

from setuptools import setup, find_packages

from immuneML.environment.Constants import Constants


def import_requirements(filename) -> list:
    with open(filename, 'r') as file:
        requirements = file.read().split("\n")
    return requirements


setup(
    name="immuneML",
    version=Constants.VERSION,
    description="immuneML is a software platform for machine learning analysis of immune receptor repertoires.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="immuneML Team",
    author_email="milenpa@student.matnat.uio.no",
    url="https://github.com/uio-bmi/immuneML",
    install_requires=["numpy", "pandas>=1,<2.0", "PyYAML>=5.3", "scikit-learn>=0.23",
                      "gensim>=4", "matplotlib>=3.1", "editdistance", "regex", "tzlocal", "airr>=1,<1.4",
                      "fishersapi", "pystache", "torch>=1.5.1", "dill>=0.3", "plotly>=4", "logomaker>=0.8",
                      "matplotlib-venn>=0.11", "scipy", "bionumpy>=0.2.31", "umap-learn"],
    extras_require={
        "TCRdist": ["tcrdist3>=0.1.6"],
        "gen_models": ['olga', 'sonnia', 'torch'],
        "ligo": ['olga', 'stitchr', 'IMGTgeneDL'],
        "KerasSequenceCNN": ["keras==2.11.0", "tensorflow==2.11.0"]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3"
    ],
    python_requires='>=3.7',
    packages=find_packages(exclude=["test", "test.*"]),
    package_data={
        'immuneML': [str(Path('IO/dataset_import/conversion/*.csv')),
                     str(Path("presentation/html/templates/*.html")),
                     str(Path("presentation/html/templates/css/*.css")),
                     str(Path('encodings/atchley_kmer_encoding/*.csv'))] +
                    [str(Path("config/default_params/") / dir.name / "*.yaml") for dir in Path("./immuneML/config/default_params/").glob("*")],
    },
    entry_points={
        'console_scripts': [
            'immune-ml = immuneML.app.ImmuneMLApp:main',
            'immune-ml-quickstart = immuneML.workflows.instructions.quickstart:main',
            'ligo = immuneML.app.LigoApp:main'
        ]
    },
)
